import numpy as np
import nifty8 as ift
from jax import numpy as jnp
from collections import namedtuple
from typing import Tuple
import itertools
from scipy.sparse import coo_matrix, save_npz, load_npz

import yaml

from ..domains.motif_space import _return_motif_categories
from ..domains.motif_production_space import (MotifProductionSpace,
                                              make_motif_production_dct,
                                              _determine_product_and_template_categories_and_ligation_spots,
                                              _production_channel_id,
                                              _valid_production_channel
                                              )

from .units import (Unit, make_unit,
                    transform_unit_to_dict, transform_dict_to_unit)

from ..utils.save import create_directory_path_if_not_already_existing

def _create_empty_motif_production_dict(motiflength : int,
        alphabet : list,
        maximum_ligation_window_length : int) -> dict:
    number_of_letters = len(alphabet)

    empty_motif_production_vector = {}
    mpd = make_motif_production_dct(
            alphabet,
            motiflength,
            maximum_ligation_window_length
            )
    for key in mpd.keys():
        empty_motif_production_vector[key] = np.zeros(mpd[key].shape)
    return empty_motif_production_vector

def MotifProductionVector(motiflength : int,
                          alphabet : list,
                          unit : Unit,
                          maximum_ligation_window_length : int
                          ) -> Tuple[object]:
    unit = make_unit(unit)
    if maximum_ligation_window_length is None:
        maximum_ligation_window_length = motiflength

    motif_production_vector_properties = {'motiflength' : motiflength,
        'alphabet' : alphabet,
        'number_of_letters' : len(alphabet),
        'unit' : unit,
        'maximum_ligation_window_length' : maximum_ligation_window_length
        }

    def makeMotifProductionVector(motif_production_vector_dct : dict):
        motif_production_vector = namedtuple('MotifProductionVector',
                ('productions',) + tuple(motif_production_vector_properties.keys()))
        productions = ift.MultiField.from_raw(
            MotifProductionSpace.make(alphabet,motiflength,maximum_ligation_window_length),
            motif_production_vector_dct,
            )
        return motif_production_vector(**{**{'productions' : productions}, **motif_production_vector_properties})
    return makeMotifProductionVector

def isinstance_motifproductionvector(obj) -> bool:
    is_motif_production_vector = True
    keys = ['motiflength','alphabet','number_of_letters','unit','maximum_ligation_window_length','productions']
    for key in obj._asdict().keys():
        if key not in keys:
            print('Not a MotifProductionVector, missing key: {}.'.format(key))
            return False
    keys = MotifProductionSpace(
        obj.alphabet,
        obj.motiflength,
        obj.maximum_ligation_window_length
    ).keys()
    for key in obj.motifs.keys():
        if key not in keys:
            print('Not a MotifVector, missing key in motifs field: {}.'.format(key))
            return False
    is_motif_vector *= isinstance(obj, tuple)
    return bool(is_motif_production_vector)

def _motif_production_array_shape(number_of_letters : int,
        maximum_ligation_window_length : int
        ) -> tuple:
    motif_production_array_shape = (number_of_letters+1,)*(maximum_ligation_window_length-maximum_ligation_window_length//2-1)
    motif_production_array_shape += (number_of_letters,)*2
    motif_production_array_shape += (number_of_letters+1,)*(maximum_ligation_window_length//2-1)
    motif_production_array_shape += motif_production_array_shape[::-1]
    return motif_production_array_shape

def _motif_production_vector_as_array(
        motif_production_vector : MotifProductionVector,
        ) -> np.ndarray:
    """
    transforms a motif vector into a numpy-array

    Parameters:
    -----------
    motif_production_vector : MotifProductionVector

    Returns:
    --------
    motif_production_array : np.ndarray
    """
    motiflength = motif_production_vector.motiflength
    number_of_letters = motif_production_vector.number_of_letters
    maximum_ligation_window_length = motif_production_vector.maximum_ligation_window_length

    motif_categories = _return_motif_categories(motiflength)

    motif_production_array = np.zeros(_motif_production_array_shape(number_of_letters,maximum_ligation_window_length))

    if maximum_ligation_window_length < 4:
        ligation_window_lengths = np.array([maximum_ligation_window_length])
    else:
        ligation_window_lengths = np.arange(4,maximum_ligation_window_length+1)

    for ligation_window_length in ligation_window_lengths:
        product_categories, template_categories, ligation_spots = _determine_product_and_template_categories_and_ligation_spots(motiflength,
                maximum_ligation_window_length,
                ligation_window_length
                )

        for product_category, template_category, ligation_spot in itertools.product(product_categories, template_categories, ligation_spots):
            if not _valid_production_channel(product_category, template_category,
                ligation_window_length, ligation_spot,
                maximum_ligation_window_length):
                continue
            reaction_key = _production_channel_id(product_category, template_category,
                    ligation_window_length, ligation_spot)

            destination_axes, source_axes = _moved_axes(ligation_window_length,ligation_spot, maximum_ligation_window_length)
            mpa_indices = _reaction_indices(product_category, template_category,
                    ligation_window_length, ligation_spot,
                    maximum_ligation_window_length, axes_moved=False)
            np.moveaxis(motif_production_array,source_axes,destination_axes)[mpa_indices] = motif_production_vector.productions[reaction_key].val
    return motif_production_array

def save_motif_production_vector(
        archive_path : str,
        motif_production_vector : MotifProductionVector,
        file_sparse : bool = True
        ):
    create_directory_path_if_not_already_existing(archive_path)
    if file_sparse:
        save_npz(archive_path+'motif_productions',
                coo_matrix(_motif_production_vector_as_array(motif_production_vector).reshape(1,-1)))
    else:
        np.save(archive_path+'motif_productions',
                _motif_production_vector_as_array(motif_production_vector)
                )

    with open(archive_path+'properties.yaml','w') as yaml_file:
        yaml.dump({'motiflength':motif_production_vector.motiflength,
            'alphabet':motif_production_vector.alphabet,
            'unit':transform_unit_to_dict(motif_production_vector.unit),
            'maximum_ligation_window_length' : motif_production_vector.maximum_ligation_window_length
            },
            yaml_file,
            indent=4)

def load_motif_production_vector(archive_path : str,
        file_sparse : bool = True
        ) -> MotifProductionVector:
    dct_filename = archive_path+'properties.yaml'
    productions_filename = archive_path+'motif_productions'
    productions_filename += '.np'+'z'*file_sparse+'y'*(1-file_sparse)
    with open(dct_filename, 'r') as yaml_file:
        properties = yaml.safe_load(yaml_file)
    if file_sparse:
        motif_production_array = np.asarray(coo_matrix.todense(load_npz(productions_filename)))
        motif_production_array = motif_production_array.reshape(
                _motif_production_array_shape(len(properties["alphabet"]),
                    properties["maximum_ligation_window_length"])
                ) 
    else:
        motif_production_array = np.load(productions_filename)
    properties['unit'] = transform_dict_to_unit(properties['unit'])
    motif_production_vec = _array_to_motif_production_vector(motif_production_array,
            **properties
            )
    return motif_production_vec

def _array_to_motif_production_vector(motif_production_array: np.ndarray,
        motiflength : int,
        alphabet : list,
        unit : Unit,
        maximum_ligation_window_length : int
        ) -> MotifProductionVector:
    unit = make_unit(unit)
    makeMotifProductionVector = MotifProductionVector(motiflength, alphabet,
            unit, maximum_ligation_window_length)
    motif_production_vector_dct = _motif_production_array_to_dct(motif_production_array,
            motiflength, alphabet, maximum_ligation_window_length
            )
    return makeMotifProductionVector(motif_production_vector_dct)

def _motif_production_array_to_dct(motif_production_array: np.ndarray,
        motiflength : int,
        alphabet : list,
        maximum_ligation_window_length : int
        ) -> dict:
    """
    transforms a motif array into a motif vector

    Parameters:
    -----------
    motif_production_array : np.ndarray

    Returns:
    --------
    motif_production_vector : MotifProductionVector
    """
    motif_production_vector_dct = {}
    number_of_letters = len(alphabet)

    motif_categories = _return_motif_categories(motiflength)

    if maximum_ligation_window_length < 4:
        ligation_window_lengths = np.array([maximum_ligation_window_length])
    else:
        ligation_window_lengths = np.arange(4,maximum_ligation_window_length+1)

    mpd = make_motif_production_dct(
            alphabet,
            motiflength,
            maximum_ligation_window_length
            )
    for ligation_window_length in ligation_window_lengths:
        product_categories, template_categories, ligation_spots = _determine_product_and_template_categories_and_ligation_spots(motiflength,
                maximum_ligation_window_length,
                ligation_window_length
                )

        for product_category, template_category, ligation_spot in itertools.product(product_categories, template_categories, ligation_spots):
            if not _valid_production_channel(product_category, template_category,
                ligation_window_length, ligation_spot,
                maximum_ligation_window_length):
                continue
            reaction_key = _production_channel_id(product_category, template_category,
                    ligation_window_length, ligation_spot)
            mpa_indices = _reaction_indices(product_category, template_category,
                    ligation_window_length, ligation_spot,
                    maximum_ligation_window_length, axes_moved=False)

            destination_axes , source_axes= _moved_axes(ligation_window_length,ligation_spot,
                    maximum_ligation_window_length)
            motif_production_vector_dct[reaction_key] = np.moveaxis(
                    motif_production_array,
                    source_axes,
                    destination_axes)[mpa_indices]
    return motif_production_vector_dct

def _moved_axes(ligation_window_length : int,
        ligation_spot : int,
        maximum_ligation_window_length : int):
    """
    returns the axes_indixes of the overlap
    in the vector (source) and in the array (destination)
    such that the motif is not interrupted in the vector and the ligation spot
    is in the center of the ligation window for the array.
    For the array, periodic boundary conditions treat longer arrays,
    where the end of the motifs is indicated by a 0 either in the motif itself
    or its hybridized partner.
    """
    ligation_spot_relative_to_center = ligation_spot-(ligation_window_length-ligation_window_length//2-1)
    source = np.arange(min(0,ligation_spot_relative_to_center),max(0,ligation_spot_relative_to_center))
    product_source = (maximum_ligation_window_length+source)%maximum_ligation_window_length
    product_destination = (maximum_ligation_window_length-source[::-1]-1)%maximum_ligation_window_length
    template_source = maximum_ligation_window_length+product_destination
    template_destination = maximum_ligation_window_length+product_source
    source = list(product_source) + list(template_source)
    destination = list(product_destination) + list(template_destination)
    return source, destination


def _reaction_indices(product_category : str,
        template_category : str,
        ligation_window_length : int,
        ligation_spot : int,
        maximum_ligation_window_length : int,
        axes_moved : bool = True
        ) -> tuple:
    motif_categories = _return_motif_categories(maximum_ligation_window_length)#FIXME: ignore monomers
    product_length = (ligation_window_length
            -int(product_category not in motif_categories[-2:])
            -int(product_category not in motif_categories[-3:-1])
            )
    # product
    left_reactant_length = ligation_spot+int(product_category in motif_categories[-2:])
    right_reactant_length = product_length-left_reactant_length
    product_first_part_overlap_length = max(0,
            left_reactant_length-maximum_ligation_window_length+maximum_ligation_window_length//2)
    product_second_part_overlap_length = max(0,
            right_reactant_length - maximum_ligation_window_length//2)
    length_from_first_product_part = min(left_reactant_length,
            maximum_ligation_window_length-maximum_ligation_window_length//2)
    length_from_second_product_part = min(right_reactant_length,
            maximum_ligation_window_length//2)
    left_ligation_window_length = left_reactant_length + int(product_category not in motif_categories[-2:])
    right_ligation_window_length = ligation_window_length-left_ligation_window_length
    ligation_window_shift = ligation_spot-ligation_window_length+ligation_window_length//2+1
    #assert(left_ligation_window_length+product_second_part_overlap_length)
    #length_from_first_product_part-product_second_part_overlap_length-int(product_category in motif_categories[-1:])

    if axes_moved:
        mpa_indices = (slice(1,None),)*product_second_part_overlap_length
        mpa_indices += (0,)*(maximum_ligation_window_length-maximum_ligation_window_length//2-length_from_first_product_part-product_second_part_overlap_length)
        mpa_indices += (slice(1,None),)*(length_from_first_product_part-1)
        mpa_indices += (slice(None),)*2
        mpa_indices += (slice(1,None),)*(length_from_second_product_part-1)
        mpa_indices += (0,)*(maximum_ligation_window_length-len(mpa_indices)-product_first_part_overlap_length)
        mpa_indices += (slice(1,None),)*product_first_part_overlap_length
    else:
        mpa_indices = (0,)*max(0,maximum_ligation_window_length-maximum_ligation_window_length//2-left_ligation_window_length+ligation_window_shift)
        mpa_indices += (0,)*int(product_category not in motif_categories[-2:])
        mpa_indices += (slice(1,None),)*(left_reactant_length-1)
        mpa_indices += (slice(None),)*2
        mpa_indices += (slice(1,None),)*(right_reactant_length-1)
        mpa_indices += (0,)*(max(0,maximum_ligation_window_length-len(mpa_indices)))

    #template
    template_length = (ligation_window_length
            -int(template_category not in motif_categories[-2:])
            -int(template_category not in motif_categories[-3:-1])
            )
    template_second_part_length = ligation_spot+int(template_category in motif_categories[-3:-1])
    template_first_part_length = template_length-template_second_part_length
    length_from_first_template_part = min(template_first_part_length,
            maximum_ligation_window_length//2)
    length_from_second_template_part = min(template_second_part_length,
            maximum_ligation_window_length-maximum_ligation_window_length//2)
    template_first_part_overlap_length = template_first_part_length-length_from_first_template_part
    template_second_part_overlap_length = template_second_part_length-length_from_second_template_part
    if ((product_second_part_overlap_length
            +length_from_first_product_part)
            >=(maximum_ligation_window_length
            -maximum_ligation_window_length//2
            +int((product_second_part_overlap_length==0)
                or (template_second_part_length<length_from_first_product_part)))
            ):
        raise ValueError('Expected: '
                + str(product_second_part_overlap_length)+'+'
                + str(length_from_first_product_part)+'<'
                + str(maximum_ligation_window_length)+'-'
                + str(maximum_ligation_window_length//2)+'+int('
                + str(product_second_part_overlap_length)+'==0 or '
                + str(template_second_part_length) + '<'
                + str(length_from_first_product_part) +')'
                + _production_channel_id(product_category, template_category,
                    ligation_window_length, ligation_spot)
                )
    if (product_first_part_overlap_length
            +length_from_second_product_part
            >=
            maximum_ligation_window_length//2
            +int(product_first_part_overlap_length==0 or length_from_first_template_part<length_from_second_product_part)):
        raise ValueError(str(product_first_part_overlap_length)+'+'
            +str(length_from_second_product_part) +'>='
            +str(maximum_ligation_window_length) + '//2 + int('
            +str(product_first_part_overlap_length) + '==0 or '
            +str(length_from_first_template_part) +'<'
            +str(length_from_second_product_part)+'))')
    if axes_moved:
        mpa_indices += (slice(1,None),)*template_second_part_overlap_length
        mpa_indices += (0,)*(maximum_ligation_window_length//2-length_from_first_template_part-template_second_part_overlap_length)
        mpa_indices += (slice(1,None),)*(length_from_first_template_part-1)
        mpa_indices += (slice(None),)*2
        mpa_indices += (slice(1,None),)*(length_from_second_template_part-1)
        mpa_indices += (0,)*(2*maximum_ligation_window_length-len(mpa_indices)-template_first_part_overlap_length)
        mpa_indices += (slice(1,None),)*template_first_part_overlap_length
    else:
        mpa_indices += (0,)*max(0,maximum_ligation_window_length//2-right_ligation_window_length-ligation_window_shift)
        mpa_indices += (0,)*int(template_category not in motif_categories[-2:])
        mpa_indices += (slice(1,None),)*(template_first_part_length-1)
        mpa_indices += (slice(None),)*2
        mpa_indices += (slice(1,None),)*(template_second_part_length-1)
        mpa_indices += (0,)*(2*maximum_ligation_window_length-len(mpa_indices))
    return mpa_indices



from .hamming_space import HammingSpace
from .motif_space import MotifSpace
from .motif_space import _return_motif_categories, _motif_categories
from .motif_hyphen_space import MotifHyphenSpace, make_hyphen_dct
from .motif_breakage_space import MotifBreakageSpace
from .time_space import TimeSpace
from .time_space import isinstance_timespace, are_compatible_timespaces
from .trajectory_space import TrajectorySpace

from .extentive_motif_breakage_space import makeExtentiveMotifBreakageSpace

from .motif_production_space import MotifProductionSpace
from .motif_production_space import make_motif_production_dct, _production_channel_id
from .motif_production_space import _return_motif_production_categories



import numpy as np
from typing import Union
from .config import symbol_config as read_symbol_config
from ..obj.motif_trajectory import MotifTrajectory
from ..obj.motif_vector import MotifVector, are_compatible_motif_vectors
from ..obj.times_vector import TimesVector
from ..obj.units import Unit
from warnings import warn

def strand_motifs_trajectory(filepaths : list,
                             alphabet : list,
                             motiflength : int = 4,
                             times_unit : Unit = None,
                             skiprows : int =2
                             ) -> MotifTrajectory:
    """
    reads from the complexes.txt of the RNAReactor simulation output and returns corresponding
    concentration vectors in motif space.

    PARAMETERS:
    -----------
    filepaths : string or list of strings
        Output file of the RNAReactor simulation
    skiprow : int, optional
        Skip the first `skiprow` lines, including comments
        when reading the file;
        default : 2

    RETURN:
    -------
    strand_motifs_trajectory : MotifTrajectory
    """
    if filepaths is str:
        filepaths = [filepaths,]
    if not isinstance(filepaths, list):
        raise ValueError("filepaths needs to be list.")
    motif_vectors = []
    times = []
    if times_unit is None:
        times_unit = read_symbol_config('time', unitformat=True)

    for filepath in filepaths:
        _, current_times, sequence_trajectory = steps_and_times_and_sequence_trajectory_from_complexes_txt(
                filepath, skiprows=skiprows)
        if len(current_times) == 0:
            continue
        motif_vectors += _transform_sequence_trajectory_into_motif_vector_list(sequence_trajectory, alphabet, motiflength)
        if (len(times)>1):
            if not are_compatible_motif_vectors(motif_vectors[0],motif_vectors[-1]):
                raise ValueError("Non compatible motif vectors.")
            if (times[-1]>current_times[0]):
                warn("Times of initialized motif trajectory will not be chronologically.")
        times += list(current_times)
    times = TimesVector(times,times_unit)
    return MotifTrajectory(motif_vectors,times)

def steps_and_times_and_complexes_from_complexes_txt(filepath : str,
        skiprows : int = 2,
        ) -> Union[np.array, np.array, list]:
    """
    Parameters:
    -----------
    filepath : str,
    skiprows : int = 2,
        skip the first <skiprows> lines of the complex.txt file

    Returns:
    --------
    steps : nd-array with dtype int
    total_physical_time : nd-array with dtype np.float64
    complexes : list
        with every element of list is the list of complexes at that time
        the list of complexes at given time is again a list of the format
        complexes[time_index][complex_index]=[number_of_complex : int, structure_of_complex : str]
    """
    if len(list(open(filepath))) == 0:
        warn("complexes.txt file is empty")
        nl = np.empty((0,3), dtype=str)
    else:
        nl = np.loadtxt(filepath, skiprows=skiprows, dtype=str, ndmin = 2)
    steps = np.array(nl[:,0], dtype = int)
    total_physical_time = np.array(nl[:,1], dtype = np.float64)
    number_and_structure_of_complexes = nl[:,2]
    from json import loads
    number_and_structure_of_complexes = [
            loads(number_and_structure_of_complexes[ii].replace('None','"none"'))
            for ii in range(len(number_and_structure_of_complexes))
            ]
    return steps, total_physical_time, number_and_structure_of_complexes

def steps_and_times_and_sequence_trajectory_from_complexes_txt(filepath : str,
        skiprows : int = 2,
        ) -> Union[np.array, np.array, list]:
    """
    Parameters:
    -----------
    filepath : str,
    skiprows : int = 2,
        skip the first <skiprows> lines of the complex.txt file

    Returns:
    --------
    steps : nd-array with dtype int
    total_physical_time : nd-array with dtype np.float64
    complexes : list
        with every element of list is the list of complexes at that time
        the list of complexes at given time is again a list of the format
        complexes[time_index][complex_index]=[number_of_complex : int, structure_of_complex : str]
    """
    steps, total_physical_time, number_and_structure_of_complexes = steps_and_times_and_complexes_from_complexes_txt(filepath,skiprows=skiprows)
    sequence_number_trajectory = [{} for ii in range(len(total_physical_time))]
    for time_idx in range(len(total_physical_time)):
        current_number_and_structure_of_complexes = number_and_structure_of_complexes[time_idx]
        for complex_index in range(len(current_number_and_structure_of_complexes)):
            number_of_complex = current_number_and_structure_of_complexes[complex_index][0]
            structure_of_complex = current_number_and_structure_of_complexes[complex_index][1]
            upper_strands_as_continuous_string, lower_strands_as_continuous_string  = _extract_upper_and_lower_strands_as_continuous_strings_from_structure_of_complex(structure_of_complex)
            upper_separate_strand_sequences = _cut_strands_as_continuous_string_into_separate_strand_sequences(upper_strands_as_continuous_string)
            lower_separate_strand_sequences = _cut_strands_as_continuous_string_into_separate_strand_sequences(lower_strands_as_continuous_string)
            for upper_separate_strand_sequence in list(upper_separate_strand_sequences):
                if upper_separate_strand_sequence in sequence_number_trajectory[time_idx].keys():
                    sequence_number_trajectory[time_idx][upper_separate_strand_sequence] += number_of_complex
                else:
                    sequence_number_trajectory[time_idx][upper_separate_strand_sequence] = number_of_complex
            for lower_separate_strand_sequence in lower_separate_strand_sequences:
                if lower_separate_strand_sequence in sequence_number_trajectory[time_idx].keys():
                    sequence_number_trajectory[time_idx][lower_separate_strand_sequence] += number_of_complex
                else:
                    sequence_number_trajectory[time_idx][lower_separate_strand_sequence] = number_of_complex
    return steps, total_physical_time, sequence_number_trajectory

def _reverse_segments(list_of_strand_segments):
    """
    reverses a strand
    """
    return [letter[::-1] for letter in list_of_strand_segments[::-1]]

def _extract_upper_and_lower_segments(structure_of_complex : list) -> Union[np.array,np.array]:
    list_of_upper_segments = np.array(structure_of_complex)[:,0]
    list_of_lower_segments =  np.array(structure_of_complex)[:,1]
    return list_of_upper_segments, list_of_lower_segments 

def _transform_list_of_segments_to_string(list_of_segments : list,
        list_of_strings_that_are_replaced : list = [('|',0),('-',''),('X',0),('x',0),('none',''),('enon',''),('00','0')]
        ) -> str:
    """
    transforms a list_of_segments into strands as continuous string
    """
    for ii in range(len(list_of_segments)):
        list_of_segments[ii] = list_of_segments[ii].replace(' ','').lstrip('5').rstrip('3')
    strands_as_continuous_string = ''.join(list_of_segments)
    for string_replace_tuple in list_of_strings_that_are_replaced:
        strands_as_continuous_string = strands_as_continuous_string.replace(str(string_replace_tuple[0]),str(string_replace_tuple[1]))
    if len(strands_as_continuous_string)>0:
        if strands_as_continuous_string[0]!='0':
            strands_as_continuous_string = '0'+strands_as_continuous_string
        if strands_as_continuous_string[-1]!='0':
            strands_as_continuous_string = strands_as_continuous_string +'0'
    return strands_as_continuous_string

def _extract_upper_and_lower_strands_as_continuous_strings_from_structure_of_complex(structure_of_complex : list) -> Union[str,str]:
    """
    takes structure_of_complex from the RNA Strand Reactor output
    and returns the upper strands sequences and the lower strands sequences.

    Parameters:
    ---------- 
    structure_of_complex : list or numpy.array

    Returns:
    --------
    upper_strands_sequences : list of str
    lower_strands_sequences : list of str
    """
    # segments are continuous parts of a complex without the end or beginning of a strand
    list_of_upper_segments, list_of_lower_segments = _extract_upper_and_lower_segments(structure_of_complex)
    list_of_reversed_lower_segments = _reverse_segments(list_of_lower_segments)
    upper_strands_as_continuous_string = _transform_list_of_segments_to_string(list_of_upper_segments)
    lower_strands_as_continuous_string = _transform_list_of_segments_to_string(list_of_reversed_lower_segments)
    return upper_strands_as_continuous_string, lower_strands_as_continuous_string

def _cut_strands_as_continuous_string_into_separate_strand_sequences(strands_as_continuous_string : str) -> list:
    """
    cut the upper strand of a complex into its single continuous strands
    """
    strands_as_continuous_string_array = np.array(list(strands_as_continuous_string))
    indices_of_empty_spots = np.where(strands_as_continuous_string_array=='0')[0]
    separate_strand_sequences = [[]]*(indices_of_empty_spots.size-1)
    for ii in range(indices_of_empty_spots.size-1):
        separate_strand_sequences[ii] = strands_as_continuous_string[indices_of_empty_spots[ii]+1:(indices_of_empty_spots[ii+1])]
    return separate_strand_sequences

def _transform_sequence_trajectory_into_motif_vector_list(
        sequence_trajectory : list,
        alphabet : list,
        motiflength : int) -> list:
    """
    Returns:
    --------
    motif_vectors : list
        list of MotifVectors
    """
    motif_vectors = [[]]*len(sequence_trajectory)
    for time_index in range(len(sequence_trajectory)):
        sequence_vector = sequence_trajectory[time_index]
        motif_vectors[time_index] = _transform_sequence_vector_into_motif_vector(sequence_vector,
                alphabet,
                motiflength)
    return motif_vectors

def _translate_letters_to_numbers(alphabet:list,
        zero_is_a_letter : bool = False):
    """
    dictionary to translate letters into numbers
    """
    if zero_is_a_letter:
        dct = { '0' : 0 }
    else:
        dct = {}
    for ii in range(len(alphabet)):
        dct[alphabet[ii]] = ii+zero_is_a_letter
    return dct

def _transform_motif_string_to_index_tuple(motif_string : str,
        alphabet : list) -> tuple:
    motif_as_tuple_of_letters = tuple(motif_string)
    """
    transforms a motif e.g. ('0','A','T','C') to corresponding indices, e.g. (0,1,3,2)

    PARAMETERS:
    -----------
    motif : array of letters

    RETURNS:
    --------
    indices
    """
    transdict = _translate_letters_to_numbers(alphabet)
    return tuple(transdict[letter] for letter in motif_as_tuple_of_letters)

def _transform_sequence_vector_into_motif_vector(sequence_vector : dict,
        alphabet : list,
        motiflength : int) -> MotifVector:
    from ..obj.motif_vector import _create_empty_motif_vector_dct
    motif_vector_dct = _create_empty_motif_vector_dct(motiflength,alphabet=alphabet)
    from ..domains.motif_space import _motif_categories
    motif_categories = _motif_categories()
    for sequence in sequence_vector.keys():
        strandlength = len(sequence)
        if (strandlength < (motiflength-1)):
            motif_category = motif_categories[0].format(strandlength)
            motif_index_tuple = _transform_motif_string_to_index_tuple(sequence,alphabet)
            motif_vector_dct[motif_category][motif_index_tuple] += sequence_vector[sequence]
        else:
            occupation_number = sequence_vector[sequence]
            # add beginning
            motif_category = motif_categories[-3]
            motif_string = sequence[:(motiflength-1)]
            motif_index_tuple = _transform_motif_string_to_index_tuple(motif_string, alphabet)
            motif_vector_dct[motif_category][motif_index_tuple] += occupation_number
            # add ending
            motif_category = motif_categories[-1]
            motif_string = sequence[-(motiflength-1):]
            motif_index_tuple = _transform_motif_string_to_index_tuple(motif_string, alphabet)
            motif_vector_dct[motif_category][motif_index_tuple] += occupation_number
            # add continuations
            motif_category = motif_categories[-2]
            for motif_index in range(len(sequence)-motiflength+1):
                motif_string = sequence[motif_index:motif_index+motiflength]
                motif_index_tuple = _transform_motif_string_to_index_tuple(motif_string, alphabet)
                motif_vector_dct[motif_category][motif_index_tuple] += occupation_number
    motif_vector = MotifVector(motiflength,alphabet,'1')
    return motif_vector(motif_vector_dct)



import nifty8 as ift
import numpy as np

from functools import reduce

from ..obj.units import make_unit, Unit

def TimeSpace(times : np.ndarray,
              units : Unit):
    times = np.asarray(times)
    if isinstance(times,(float,np.float64,int)) or times.shape[0]<=1:
        return UnstructuredTimeDomain(1, make_unit(units))
    if isinstance(times,list) and not isinstance(times[0],(float,int)):
        raise ValueError("times needs to be 1d-array")
    elif len(times.shape) != 1:
        raise ValueError("times needs to be 1d-array")
    distances = np.diff(times)
    if np.all(distances==distances[0]):
        time_space = TimeRGSpace(times.shape, make_unit(units), distances=distances[0])
    else:
        time_space = UnstructuredTimeDomain(times.shape, make_unit(units))
    return time_space


def isinstance_timespace(obj)->bool:
    if isinstance(obj,ift.DomainTuple):
        obj = obj[0]
    if not 'units' in dir(obj):
        print("Units not specified")
        return False
    if not (isinstance(obj, TimeRGSpace) or isinstance(obj,UnstructuredTimeDomain)):
        print('not RGSpace or UnstructuredDomain')
        return False
    return True

def are_compatible_timespaces(ts1 : object,
                              ts2 : object
                              )->bool:
    if isinstance(ts1,ift.DomainTuple):
        ts1 = ts1[0]
    if isinstance(ts2,ift.DomainTuple):
        ts2 = ts2[0]
    if not isinstance_timespace(ts1):
        print('Object is not a TimeSpace')
        return False
    if not isinstance_timespace(ts2):
        print('Object is not a TimeSpace')
        return False
    if ts1.units != ts2.units:
        print('times vectors have different units: {}'.format(ts1.units) + ' and {}'.format(ts2.units))
        return False
    if ts1!=ts2:
        print('DomainMismatch')
        return False
    return True

class TimeRGSpace(ift.RGSpace):
    """
    RGSpace for time
    """
    _needed_for_hash = ['_rdistances','_shape','_harmonic']
    def __init__(self, shape, units, distances):
        self._units = make_unit(units)
        self._harmonic = False
        if np.isscalar(shape):
            shape = (shape,)
        self._shape = tuple(int(i) for i in shape)
        if min(self._shape) < 0:
            raise ValueError('Negative number of pixels encountered')
        #
        self._rdistances = (float(distances),) * len(self._shape)
        self._hdistances = tuple(
            1. / (np.array(self.shape)*np.array(self._rdistances)))
        #
        self._dvol = float(reduce(lambda x, y: x*y, self.distances))
        self._size = int(reduce(lambda x, y: x*y, self._shape))

    @property
    def units(self):
        return self._units

class UnstructuredTimeDomain(ift.UnstructuredDomain):
    """
    UnstructuredDomain for time
    """
    _needed_for_hash = ['_shape']
    def __init__(self, shape, units):
        self._units = make_unit(units)
        try:
            self._shape = tuple([int(i) for i in shape])
        except TypeError:
            self._shape = (int(shape), )

    @property
    def units(self):
        return self._units



import numpy as np
from nifty8 import RGSpace
from functools import reduce

class HammingSpace(RGSpace):
    """Represents a Hamming Space

    Parameters:
    -----------
    alphabet : list of strings
        alphabet of the Hamming space.
    wordlength : int
        Number of letters in each word.
    """
    def __init__(self, alphabet, wordlength):
        self._harmonic = False
        if not isinstance(wordlength,(int,np.int64)):
            raise ValueError('Non-integer wordlength encountered')
        if not isinstance(alphabet,list):
            raise ValueError('Alphabet needs to be a list')
        self._alphabet = alphabet
        self._wordlength = wordlength
        self._shape = (len(self.alphabet),)*self.wordlength
        #
        distances = 1.
        self._rdistances = (float(distances),) * len(self._shape)
        self._hdistances = tuple(
            1. / (np.array(self.shape)*np.array(self._rdistances)))
        #
        self._dvol = 1.
        self._size = int(reduce(lambda x, y: x*y, self._shape))

    def __repr__(self):
        return ("HammingSpace(alphabet={}, wordlength={})"
                .format(self.alphabet, self.wordlength))

    @property
    def alphabet(self):
        return self._alphabet

    @property
    def wordlength(self):
        return self._wordlength



import numpy as np

from .effective_ligation_rates_from_parameters import separate_hybridization_and_ligation_parameters

from ..obj.motif_vector import MotifVector

def onset_of_growth(
        total_mass,
        initial_dimer_concentration,
        extension_rate_constant_1_1_2,
        extension_rate_constant_1_2_2,
        cleavage_rate_constant,
    ) -> float:
    return (np.log(total_mass**2*extension_rate_constant_1_1_2-cleavage_rate_constant)-np.log(total_mass*extension_rate_constant_1_2_2*initial_dimer_concentration))/(total_mass**2*extension_rate_constant_1_1_2-cleavage_rate_constant)

def discretized_onset_of_growth(
        total_mass,
        initial_dimer_concentration,
        extension_rate_constant_1_1_2,
        cleavage_rate_constant,
        concentration_of_a_single_particle,
    ) -> float:
    return 1./(total_mass**2*extension_rate_constant_1_1_2-cleavage_rate_constant)*np.log(1.+concentration_of_a_single_particle/initial_dimer_concentration)

def onset_of_growth_from_strand_trajectory_parameters(
        initial_motif_numbers_vector : MotifVector,
        strand_reactor_parameters : dict
    ) -> float:
    ligation_rate_constant = np.exp(strand_reactor_parameters['dG_4_2Match_mean']*strand_reactor_parameters['l_critical'])
    gamma2nc = strand_reactor_parameters['dG_4_0Match']
    gamma1nc = strand_reactor_parameters['dG_4_1Match']
    gammacom = strand_reactor_parameters['dG_4_2Match_mean']
    inverse_dissociation_constant = 1./4.*(
        np.exp(-gamma2nc)
        +2*np.exp(-gamma1nc)
        +np.exp(-gammacom)
    )
    k112 = ligation_rate_constant*inverse_dissociation_constant

    epsilon1nc = strand_reactor_parameters['dG_3_0Match']
    epsiloncom = strand_reactor_parameters['dG_3_1Match_mean']
    inverse_dissociation_constant = 1./4.*(
        np.exp(gamma2nc+epsilon1nc)
        +np.exp(gamma1nc+epsilon1nc)
        +np.exp(gamma1nc+epsiloncom)
        +np.exp(gammacom+epsiloncom)
        )
    k122 = ligation_rate_constant*inverse_dissociation_constant

    concentration_of_a_single_particle = strand_reactor_parameters['c_ref']/initial_motif_numbers_vector.motifs.val['length1strand'][0]
    total_mass = np.sum(initial_motif_numbers_vector.motifs.val['length1strand'])*concentration_of_a_single_particle
    initial_dimer_concentration = np.sum(initial_motif_numbers_vector.motifs.val['length2strand'])*concentration_of_a_single_particle
    return onset_of_growth(
        total_mass,
        initial_dimer_concentration,
        k112,
        k122,
        strand_reactor_parameters['r_delig']
    )



from ..obj.motif_trajectory import MotifTrajectory
from ..obj.motif_trajectory import _motif_trajectory_as_array
from ..obj.motif_trajectory import _array_to_motif_trajectory
from ..obj.units import make_unit, Unit

def motif_concentration_trajectory_from_motif_number_trajectory(
        motif_number_trajectory : MotifTrajectory,
        c_ref : float,
        unit : Unit = make_unit('mol')/make_unit('L')
        ) -> MotifTrajectory:
    single_particle_concentration = c_ref/motif_number_trajectory.motifs['length1strand'].val[0][0]

    motif_number_trajectory_array, _ = _motif_trajectory_as_array(
            motif_number_trajectory
            )
    motif_concentration_trajectory_array = single_particle_concentration*motif_number_trajectory_array
    return _array_to_motif_trajectory(
            motif_concentration_trajectory_array,
            motif_number_trajectory.times,
            motiflength  = motif_number_trajectory.motiflength,
            alphabet = motif_number_trajectory.alphabet,
            unit = unit
            )



import morsaik as kdi
import numpy as np
import pytest

from numpy.testing import assert_equal

def test_motif_production_trajectory_ensemble():
    alphabet = ['a','b','c']
    motiflength = 5
    unit = kdi.make_unit('mol')/kdi.make_unit('L')
    times = kdi.TimesVector(np.arange(3), 'ms')
    maximum_ligation_window_length = motiflength

    motif_categories = kdi.domains._return_motif_categories(motiflength=motiflength)
    motif_production_dct = kdi._create_empty_motif_production_dict(motiflength,
                                                                     alphabet,
                                                                     maximum_ligation_window_length
                                                                     )

    motif_production_vector = kdi.MotifProductionVector(motiflength,
                                                        alphabet,
                                                        unit,
                                                        maximum_ligation_window_length
                                                        )
    motif_production_vector = motif_production_vector(motif_production_dct)
    motif_production_vectors = [motif_production_vector,]*times.shape[0]
    mt = kdi.MotifProductionTrajectory(motif_production_vectors,times)

    mts = [mt,mt]

    mte = kdi.MotifProductionTrajectoryEnsemble(mts)

    assert_equal(mte.trajectories[0].times.val,mt.times.val)
    assert_equal(mte.trajectories[0].times.domain,mt.times.domain)
    assert_equal(mte.trajectories[1].times.val,mt.times.val)
    assert_equal(mte.trajectories[1].times.domain,mt.times.domain)
    assert_equal(mte.alphabet, mt.alphabet)
    assert_equal(mte.number_of_letters, mt.number_of_letters)
    assert(mte.motiflength==mt.motiflength)
    assert(mte.unit==mt.unit)
    for trajectory in mte.trajectories:
        for ii in range(len(trajectory.times.val)):
            assert_equal(trajectory.productions.val,
                    mt.productions.val
                    )

    assert(kdi.isinstance_motifproductiontrajectoryensemble(mte))
    print('mte')
    print(mte)
    assert(kdi.are_compatible_motif_production_trajectory_ensembles(mte,mte))

def test_save_and_load_motif_production_trajectory_ensemble():
    motiflengths = [4,]#[2,3,4,5,6,10]

    for motiflength in motiflengths:
        strand_trajectory_id='9999_99_99__99_99_99'
        archive_path = kdi.utils.create_trajectory_ensemble_path(
                strand_trajectory_id=strand_trajectory_id,
                param_file_no = 0,
                motiflength=motiflength
                )

        motif_production_trajectory = kdi.get.strand_motifs_productions_trajectory(motiflength, strand_trajectory_id)
        motif_production_trajectories = [motif_production_trajectory,]*2
        mte = kdi.MotifProductionTrajectoryEnsemble(motif_production_trajectories)

        kdi.save_motif_production_trajectory_ensemble(archive_path, mte)
        actual_mte = kdi.load_motif_production_trajectory_ensemble(archive_path)
        assert(mte.motiflength==actual_mte.motiflength)
        assert(mte.alphabet==actual_mte.alphabet)
        assert(mte.maximum_ligation_window_length==actual_mte.maximum_ligation_window_length)
        assert(mte.number_of_letters==actual_mte.number_of_letters)
        assert(mte.unit==actual_mte.unit)
        for trajectory_index in range(len(actual_mte.trajectories)):
            actual = kdi._motif_production_trajectory_as_array(actual_mte.trajectories[trajectory_index])
            assert_equal(actual,
                    kdi._motif_production_trajectory_as_array(motif_production_trajectories[trajectory_index])
                    )



import matplotlib.pyplot as plt
from os import makedirs

from .zebraness import system_level_motif_zebraness as plot_system_level_motif_zebraness
from .entropy import motif_entropy as plot_motif_entropy

def motif_trajectory_analysis(
        motif_trajectory_ensembles : list,
        plot_instructions : dict,
        plot_parameters_list : list,
        plotting_time_windows : list = None,
        plotpath : str = './Plots/'
    ):
    """
    Meta function to plot all analysis plots as specified in the plotting constructions.
    """
    makedirs(plotpath, exist_ok=True)
    if plot_instructions['plot_system_level_motif_zebraness']:
        ii = 0
        for motif_trajectory_ensemble in  motif_trajectory_ensembles:
            plot_parameters = plot_parameters_list[ii]
            plot_system_level_motif_zebraness(
                motif_trajectory_ensemble,
                plotting_time_windows = plotting_time_windows,
                plot_parameters = plot_parameters,
                plotpath = plotpath
            )
    if plot_instructions['plot_motif_entropy']:
        ii = 0
        for motif_trajectory_ensemble in  motif_trajectory_ensembles:
            plot_parameters = plot_parameters_list[ii]
            plot_motif_entropy(
                motif_trajectory_ensemble,
                linestyle = plot_parameters['linestyle'],
                color = plot_parameters['color'],
            )
            plt.close('all')
            ii+=1



from ..obj.motif_vector import MotifVector
from ..obj.motif_trajectory import MotifTrajectory

from ..domains.motif_space import _motif_categories

import jax.numpy as jnp

def total_mass(
        motif_vector : MotifVector
        ) -> jnp.ndarray:
    """
    Parameters:
    -----------
    motif_vector : MotifVector

    Returns:
    --------
    total_mass : jnp.ndarray
    """
    motif_categories = _motif_categories()
    total_mass = 0
    for strandlength in range(1,motif_vector.motiflength-1):
        total_mass += strandlength*jnp.sum(motif_vector.motifs.val[motif_categories[0].format(strandlength)])
    total_mass += jnp.sum(motif_vector.motifs.val[motif_categories[-3]].flatten())
    total_mass += jnp.sum(motif_vector.motifs.val[motif_categories[-1]].flatten())*(motif_vector.motiflength-2)
    total_mass += jnp.sum(motif_vector.motifs.val[motif_categories[-2]].flatten())
    return total_mass

def total_mass_of_motif_concentration_trajectory_array(
        motif_concentration_trajectory_array : jnp.ndarray,
        number_of_letters : int = 4,
        motiflength : int = 4
        ) -> jnp.ndarray:
    strandmasses = jnp.concatenate([jnp.array([strandlength,]*(number_of_letters**strandlength)) for strandlength in range(1,motiflength-1)] + [jnp.array([1,]*(number_of_letters**(motiflength-1)))] + [jnp.array([1,]*(number_of_letters**(motiflength)))] + [jnp.array([motiflength-2,]*(number_of_letters**(motiflength-1)))])
    return jnp.vecdot(strandmasses[None], motif_concentration_trajectory_array, axis=1)

def total_mass_trajectory(
        motif_trajectory : MotifTrajectory
        ) -> jnp.ndarray:
    motif_categories = _motif_categories()
    total_mass_trajectory = jnp.zeros(motif_trajectory.times.size)
    for strandlength in range(1,motif_trajectory.motiflength-1):
        strand_mass_trajectory = strandlength*jnp.sum(
                motif_trajectory.motifs.val[motif_categories[0].format(strandlength)].reshape(motif_trajectory.times.size,-1),
                axis=-1)
        total_mass_trajectory = total_mass_trajectory.at[:].add(strand_mass_trajectory)
    # beginnings
    strand_mass_trajectory = jnp.sum(
            motif_trajectory.motifs.val[motif_categories[-3]].reshape(motif_trajectory.times.size,-1),
            axis=-1
            )
    total_mass_trajectory = total_mass_trajectory.at[:].add(strand_mass_trajectory)
    # continuations
    strand_mass_trajectory = jnp.sum(
            motif_trajectory.motifs.val[motif_categories[-2]].reshape(motif_trajectory.times.size,-1),
            axis=-1
            )
    total_mass_trajectory = total_mass_trajectory.at[:].add(strand_mass_trajectory)
    # endings
    strand_mass_trajectory = jnp.sum(
            motif_trajectory.motifs.val[motif_categories[-1]].reshape(motif_trajectory.times.size,-1),
            axis=-1
            )
    total_mass_trajectory = total_mass_trajectory.at[:].add((motif_trajectory.motiflength-2)*strand_mass_trajectory)
    return total_mass_trajectory



from .motif_vector import MotifVector
from .units import make_unit

def MotifNumberVector(motiflength : int,
                      alphabet : str):
    unit = make_unit('particles')
    return MotifVector(motiflength, alphabet, unit)



import pytest
import morsaik as kdi
import numpy as np
from numpy.testing import assert_equal

def test_motif_breakage_trajectory_ensemble():
    motiflengths = [2,3,4,5,6,9,10]
    for motiflength in motiflengths:
        alphabet = ['a','b','c']
        unit = kdi.make_unit('')
        times = kdi.TimesVector(np.arange(3), 'ms')

        motif_categories = kdi.domains._return_motif_categories(motiflength=motiflength)
        motif_breakage_dct = kdi._create_empty_motif_breakage_dct(motiflength,
                                                                        alphabet,
                                                                        )

        motif_breakage_vector = kdi.MotifBreakageVector(motiflength,
                                                            alphabet,
                                                            unit,
                                                            )
        motif_breakage_vector = motif_breakage_vector(motif_breakage_dct)
        motif_breakage_vectors = [motif_breakage_vector,]*times.shape[0]
        mbt = kdi.MotifBreakageTrajectory(motif_breakage_vectors,times)

        mbts = [mbt,mbt]

        mbte = kdi.MotifBreakageTrajectoryEnsemble(mbts)

        assert_equal(mbte.trajectories[0].times.val,mbt.times.val)
        assert_equal(mbte.trajectories[0].times.domain,mbt.times.domain)
        assert_equal(mbte.trajectories[1].times.val,mbt.times.val)
        assert_equal(mbte.trajectories[1].times.domain,mbt.times.domain)
        assert_equal(mbte.alphabet, mbt.alphabet)
        assert_equal(mbte.number_of_letters, mbt.number_of_letters)
        assert(mbte.motiflength==mbt.motiflength)
        assert(mbte.unit==mbt.unit)
        for trajectory in mbte.trajectories:
            for ii in range(len(trajectory.times.val)):
                assert_equal(trajectory.breakages.val,
                        mbt.breakages.val
                        )

        assert(kdi.isinstance_motifbreakagetrajectoryensemble(mbte))
        print('mbte')
        print(mbte)
        assert(kdi.are_compatible_motif_breakage_trajectory_ensembles(mbte,mbte))

@pytest.mark.skip
def test_save_and_load_motif_breakage_trajectory_ensemble():
    alphabet = ['A','T']
    motiflengths = [4,]#[2,3,4,5,6,10]
    complements = [1,0]
    unit = kdi.make_unit('mol')/kdi.make_unit('L')
    times = kdi.TimesVector(np.arange(3), kdi.read.symbol_config('time', unitformat=True))

    for motiflength in motiflengths:
        strand_trajectory_id='9999_99_99__99_99_99'
        archive_path = kdi.utils.create_trajectory_ensemble_path(
                strand_trajectory_id=strand_trajectory_id,
                param_file_no = 0,
                motiflength=motiflength
                )

        motif_breakage_vector = kdi.get.motif_breakage_rate_constants_from_strand_reactor_parameters(strand_trajectory_id, motiflength, complements)
        motif_breakage_vectors = [motif_breakage_vector,]*times.shape[0]
        motif_breakage_trajectory = kdi.MotifBreakageTrajectory(motif_breakage_vectors,times)
        motif_breakage_trajectories = [motif_breakage_trajectory,]*2
        mbte = kdi.MotifBreakageTrajectoryEnsemble(motif_breakage_trajectories)

        kdi.save_motif_breakage_trajectory_ensemble(archive_path, mbte)
        actual_mbte = kdi.load_motif_breakage_trajectory_ensemble(archive_path)
        assert(mbte.motiflength==actual_mbte.motiflength)
        assert(mbte.alphabet==actual_mbte.alphabet)
        assert(mbte.number_of_letters==actual_mbte.number_of_letters)
        assert(mbte.unit==actual_mbte.unit)
        '''
        for trajectory_index in range(len(actual_mbte.trajectories)):
            actual = kdi._motif_breakage_trajectory_as_array(actual_mbte.trajectories[trajectory_index])
            assert_equal(actual,
                    kdi._motif_breakage_trajectory_as_array(motif_breakage_trajectories[trajectory_index])
                    )
        '''



import matplotlib.pyplot as plt
import numpy as np

from ..obj.motif_trajectory import _motif_trajectory_as_array
from ..obj.units import transform_unit_to_str
from .colors import _choose_color

from matplotlib.colors import LinearSegmentedColormap

def entropy(times, 
        x_t,
        linestyle = '-',
        color = 'b',
        alpha = 1.,
        label = None,
        axis = None
    ) -> None:
    if axis is None:
        axis = plt.gca()
    x_t = x_t.reshape((len(x_t),-1))
    x_total = np.sum(x_t, axis=1)
    p = x_t / x_total[:,None]
    entrpy = np.zeros(p.shape)
    entrpy[p!=0] = p[p!=0]*np.log2(p[p!=0])
    entrpy = -np.sum(entrpy, axis = 1)
    axis.plot(times.val,
            entrpy, 
            linestyle = linestyle,
            color = color,
            alpha = alpha,
            label = label
        )
    axis.set_xlabel('Time ' + transform_unit_to_str(times.domain[0].units))
    axis.set_ylabel('Entropy [bits]')

def motif_entropy(
        motif_trajectory_ensemble,
        linestyle = '-',
        color = 'b',
        alpha = 1.,
        label = None,
        axis = None
    ) -> None:
    ii=0
    for motif_trajectory in motif_trajectory_ensemble.trajectories:
        colr = _choose_color(color, ii/np.max([1,len(motif_trajectory_ensemble.trajectories)-1]))
        times = motif_trajectory.times
        x_t, _ = _motif_trajectory_as_array(motif_trajectory)
        entropy(times,
                x_t,
                linestyle = linestyle,
                color = colr,
                alpha = alpha,
                label = label,
                axis = axis
        )
        ii += 1



import units
from warnings import warn
from typing import Union

units_dct = {} #_load_units()
Unit = units.quantity.Quantity

def make_unit(unit : Union[units.unit, Unit, str], prefactor : float = 1.):
    """
    checks if unit is a units object or quantity and multiplies (weights) it with the quantity value,
    if it is a string, it transforms it and returns the string as a units.quantity.Quantity
    else it raises an ValueError

    Parameters:
    -----------
    unit : units.unit or str

    Returns:
    --------
    unit : units.quantity.Quantity
    """
    if isunits_unit(unit):
        return unit(prefactor)
    elif isunit(unit):
        return prefactor*unit
    elif unit in ['', '1.'] or unit is None or unit==1.:
        warn("unit not specified, will treat it as one.")
        return 1.
    elif unit in ['1'] or unit == 1:
        warn("unit not specified, will treat it as one.")
        return 1
    elif isinstance(unit,dict):
        return transform_dict_to_unit(unit)
    elif isinstance(unit, str):
        warn("unit is string, will make it units object but this can lead to misinterpretation")
        return units.unit(unit)(prefactor)
    else:
        raise ValueError("unit {} is not a units.unit nor units.quantity.Quantity nor string".format(unit))

def isunits_unit(unit):
    return isinstance(unit, (units.LeafUnit,units.ComposedUnit, units.NamedComposedUnit))

def isunit(unit):
    return isinstance(unit, Unit)

def _load_units(symbol_config_path : str):
    """
    """
    raise NotImplementedError("load units not implemented yet. Sorry for the inconvenience.")

def transform_unit_to_dict(unit : Unit) -> dict:
    unit_dct = {}
    try:
        unit_dct['prefactor'] = unit.get_num()
    except AttributeError:
        if isinstance(unit,int) or isinstance(unit,float):
            unit_dct['prefactor'] = unit
            unit_dct['numerator_units'] = [None,]
            unit_dct['denominator_units'] = [None,]
            unit_dct['multiplier'] = 1.
            return unit_dct
        else:
            unit_dct['prefactor'] = 1.
            unit_dct['numerator_units'] = [str(unit),]
            unit_dct['denominator_units'] = [None,]
            unit_dct['multiplier'] = 1.
            return unit_dct
    unit = unit.get_unit()
    if isinstance(unit,units.LeafUnit):
        unit_dct['numerator_units'] = [unit.specifier,]
        unit_dct['denominator_units'] = [None,]
        unit_dct['multiplier'] = 1.
    else:
        unit_dct['numerator_units'] = [uu.specifier for uu in unit.orig_numer]
        unit_dct['denominator_units'] = [uu.specifier for uu in unit.orig_denom]
        unit_dct['multiplier'] = unit.orig_multiplier
    return unit_dct

def transform_dict_to_unit(unit_dct : dict) -> Unit:
    unit = 1.
    for numerator_unit in unit_dct['numerator_units']:
        unit = unit*make_unit(numerator_unit)
    for denominator_unit in unit_dct['denominator_units']:
        unit = unit/make_unit(denominator_unit)
    try:
        unit.orig_mutliplier = unit_dct['multiplier']
    except AttributeError:
        unit *= unit_dct['multiplier']
    return make_unit(unit, unit_dct['prefactor'])

def transform_unit_to_str(unit : Unit) -> str:
    unit = transform_unit_to_dict(unit)
    prefactor = unit['multiplier']*unit['prefactor']
    if prefactor == 1.:
        prefactor = ''
    else:
        prefactor = str(prefactor)
    numerator_units = ''
    for nu in unit['numerator_units']:
        numerator_units += str(nu) + ' '
    if unit['denominator_units'] == [None,]:
        unit =  numerator_units
    else:
        denominator_units = ''
        for du in unit['denominator_units']:
            denominator_units += str(du) + ' '
        unit = "\\frac\{" + numerator_units + "\}\{"+  denominator_units + "\}"
    if len(unit)*len(prefactor) > 0:
        prefactor += ' '
    print(prefactor)
    print(unit)
    return '$[' + prefactor + unit + ']$'



import numpy as np
from ..obj.motif_trajectory import MotifTrajectory
from ..domains.motif_space import _motif_categories

def mean_length(
        motif_trajectory : MotifTrajectory
    ):
    motif_groups = _motif_categories()
    mean_length = np.zeros(motif_trajectory.times.shape)
    total_number_of_strands = np.zeros(motif_trajectory.times.shape)
    for length in range(1,motif_trajectory.motiflength-1):
        strand_category = motif_groups[0].format(length)
        number_of_strands = np.sum(
            motif_trajectory.motifs.val[strand_category].reshape(motif_trajectory.times.shape+(-1,)),
            axis=-1
        )
        mean_length += length*number_of_strands
        total_number_of_strands += number_of_strands
    # beginnings_concentration
    number_of_strands = np.sum(
        motif_trajectory.motifs[motif_groups[-3]].val.reshape(motif_trajectory.times.shape+(-1,)),
        axis=-1
    )
    mean_length += number_of_strands
    total_number_of_strands += number_of_strands
    # continuations
    mean_length += np.sum(
        motif_trajectory.motifs[motif_groups[-2]].val.reshape(motif_trajectory.times.shape+(-1,)),
        axis=-1
    )
    # ends_concentration
    mean_length += (motif_trajectory.motiflength-2)*np.sum(
        motif_trajectory.motifs[motif_groups[-1]].val.reshape(motif_trajectory.times.shape+(-1,)),
        axis=-1
    )
    return mean_length/total_number_of_strands



from .colors import standard_colorbar, colorbar_example, green_colorbar, greenish_colorbar
from .entropy import entropy, motif_entropy

from .motif_concentrations import motif_trajectories

from .motif_production_rates import motif_production_rates
from .motif_trajectory_analysis import motif_trajectory_analysis

from .zebraness import zebraness, system_level_motif_zebraness



from ..obj.motif_concentration_vector import MotifConcentrationVector
from ..obj.motif_number_vector import MotifNumberVector
from ..obj.motif_vector import _motif_vector_as_array, _array_to_motif_vector_dct

def motif_concentration_vector_from_motif_number_vector(
        motif_number_vector : MotifNumberVector,
        c_ref : float
        ) -> MotifConcentrationVector:
    """
    transform a motif number vector intro motif concentration vector

    Parameters:
    -----------
    motif_number_vector : MotifNumberVector
    c_ref : float
        reference concentration of monomers with the first letter

    Returns:
    --------
    moif_concentration_vector : MotifConcentrationVector
    """
    motif_concentration_vector = MotifConcentrationVector(motif_number_vector.motiflength,motif_number_vector.alphabet)
    single_particle_concentration = c_ref/motif_number_vector.motifs['length1strand'][0]
    motif_concentration_array = single_particle_concentration*_motif_vector_as_array(motif_number_vector)
    return motif_concentration_vector(_array_to_motif_vector_dct(motif_concentration_array))



from collections import namedtuple
import numpy as  np
from numpy.testing import assert_equal
import yaml
from os.path import exists
from .motif_trajectory import are_compatible_motif_trajectories
from .motif_trajectory import save_motif_trajectory, load_motif_trajectory
from .units import transform_dict_to_unit, transform_unit_to_dict

from ..utils.save import create_directory_path_if_not_already_existing

def MotifTrajectoryEnsemble(motif_trajectories : list):
    """
    Ensemble of motif trajectories
    """
    motif_trajectory = motif_trajectories[0]
    motif_trajectory_ensemble_properties = {
            'motiflength' : motif_trajectory.motiflength,
            'alphabet' : motif_trajectory.alphabet,
            'number_of_letters' : motif_trajectory.number_of_letters,
            'unit' : motif_trajectory.unit,
            'times' : motif_trajectory.times,
            'trajectories' : motif_trajectories}
    for motif_trajectory in motif_trajectories:
        assert_equal(motif_trajectory.motiflength, motif_trajectory_ensemble_properties['motiflength'])
        assert_equal(motif_trajectory.alphabet, motif_trajectory_ensemble_properties['alphabet'])
        assert_equal(motif_trajectory.number_of_letters, motif_trajectory_ensemble_properties['number_of_letters'])
        assert motif_trajectory.unit==motif_trajectory_ensemble_properties['unit'], "Units of motif vectors need to be the same."
        assert(are_compatible_motif_trajectories(motif_trajectory,motif_trajectories[0]),
                "Motif trajectories not compatible.")
    mt = namedtuple('MotifTrajectoryEnsemble',
                tuple(motif_trajectory_ensemble_properties.keys()))
    return mt(**motif_trajectory_ensemble_properties)

def isinstance_motiftrajectoryensemble(obj) -> bool:
    is_motif_trajectory_ensemble = True
    keys = ['motiflength', 'alphabet', 'number_of_letters', 'unit', 'trajectories', 'times']
    for key in obj._asdict().keys():
        if key not in keys:
            print('Not a MotifTrajectoryEnsemble, missing key: {}.'.format(key))
            return False
    is_motif_trajectory_ensemble *= isinstance(obj, tuple)
    is_motif_trajectory_ensemble *= hasattr(obj, '_asdict')
    is_motif_trajectory_ensemble *= hasattr(obj, '_fields')
    return bool(is_motif_trajectory_ensemble)

def are_compatible_motif_trajectory_ensembles(trjs1 : MotifTrajectoryEnsemble, trjs2 : MotifTrajectoryEnsemble) -> bool:
    if not isinstance_motiftrajectoryensemble(trjs1):
        print('Object is not a MotifTrajectoryEnsemble')
        return False
    if not isinstance_motiftrajectoryensemble(trjs2):
        print('Object is not a MotifTrajectoryEnsemble')
        return False
    if are_compatible_motif_trajectories(trjs1.trajectories[0],trjs2.trajectories[0]):
        return True
    else:
        print('MotifTrajectoryEnsembles not compatible: {} mismatch.'.format(key))

def save_motif_trajectory_ensemble(archive_path : str,
        motif_trajectory_ensemble : MotifTrajectoryEnsemble
        ) -> None:
    create_directory_path_if_not_already_existing(archive_path)
    for trajectory_index in range(len(motif_trajectory_ensemble.trajectories)):
        current_path = archive_path+'motif_trajectory_{}/'.format(trajectory_index)
        create_directory_path_if_not_already_existing(current_path)
        save_motif_trajectory(current_path,
                motif_trajectory_ensemble.trajectories[trajectory_index]
                )
    with open(archive_path+'motif_trajectory_ensemble_properties.yaml','w') as yaml_file:
        yaml.dump({'motiflength':motif_trajectory_ensemble.motiflength,
            'alphabet':motif_trajectory_ensemble.alphabet,
            'unit':transform_unit_to_dict(motif_trajectory_ensemble.unit)},
            yaml_file,
            indent=4)

def load_motif_trajectory_ensemble(archive_path : str
        ) -> MotifTrajectoryEnsemble:
    dct_filename = archive_path+'motif_trajectory_ensemble_properties.yaml'
    motif_trajectories = []
    trajectory_index = 0
    while exists(archive_path+'motif_trajectory_{}/'.format(trajectory_index)):
        motif_trajectories = motif_trajectories + [load_motif_trajectory(archive_path+'motif_trajectory_{}/'.format(trajectory_index)),]
        trajectory_index += 1
    return MotifTrajectoryEnsemble(motif_trajectories)



from . import domains
from . import read
from . import utils
from . import get
from . import infer
from . import plot

from .obj.units import Unit, make_unit, transform_unit_to_dict, transform_dict_to_unit, transform_unit_to_str

from .obj.motif_vector import MotifVector, _create_empty_motif_vector_dct
from .obj.motif_vector import add_zebra_fluctuation_to_motif_vector
from .obj.motif_vector import convert_homogeneous_dimers_to_zebra_dimers, convert_dimers_to_tetramers
from .obj.motif_vector import isinstance_motifvector, are_compatible_motif_vectors
from .obj.motif_vector import _transform_sequence_array_to_motif_array
from .obj.motif_vector import _motif_vector_as_array, _array_to_motif_vector_dct
from .obj.motif_vector import _motif_indices_in_motifs_array
from .obj.motif_vector import save_motif_vector, load_motif_vector

from .obj.motif_number_vector import MotifNumberVector
from .obj.motif_concentration_vector import MotifConcentrationVector

from .obj.motif_trajectory import MotifTrajectory
from .obj.motif_trajectory import isinstance_motiftrajectory, are_compatible_motif_trajectories
from .obj.motif_trajectory import save_motif_trajectory, load_motif_trajectory
from .obj.motif_trajectory import extract_initial_motif_vector_from_motif_trajectory
from .obj.motif_trajectory import _motif_trajectory_as_array
from .obj.motif_trajectory import _array_to_motif_trajectory

from .obj.times_vector import TimesVector, are_compatible_times_vectors

from .obj.motif_trajectory_ensemble import MotifTrajectoryEnsemble
from .obj.motif_trajectory_ensemble import isinstance_motiftrajectoryensemble, are_compatible_motif_trajectory_ensembles
from .obj.motif_trajectory_ensemble import save_motif_trajectory_ensemble, load_motif_trajectory_ensemble

from .obj.motif_production_vector import MotifProductionVector, _create_empty_motif_production_dict
from .obj.motif_production_vector import (_motif_production_vector_as_array, _array_to_motif_production_vector)
from .obj.motif_production_vector import _motif_production_array_shape
from .obj.motif_production_vector import save_motif_production_vector, load_motif_production_vector

from .obj.motif_production_trajectory import MotifProductionTrajectory
from .obj.motif_production_trajectory import isinstance_motifproductiontrajectory, are_compatible_motif_production_trajectories
from .obj.motif_production_trajectory import _motif_production_trajectory_as_array
from .obj.motif_production_trajectory import save_motif_production_trajectory, load_motif_production_trajectory

from .obj.motif_production_trajectory_ensemble import MotifProductionTrajectoryEnsemble
from .obj.motif_production_trajectory_ensemble import isinstance_motifproductiontrajectoryensemble
from .obj.motif_production_trajectory_ensemble import are_compatible_motif_production_trajectory_ensembles
from .obj.motif_production_trajectory_ensemble import save_motif_production_trajectory_ensemble, load_motif_production_trajectory_ensemble

from .obj.motif_breakage_vector import MotifBreakageVector, isinstance_motifbreakagevector
from .obj.motif_breakage_vector import _array_to_motif_breakage_vector, _create_empty_motif_breakage_dct
from .obj.motif_breakage_trajectory import MotifBreakageTrajectory
from .obj.motif_breakage_trajectory import isinstance_motifbreakagetrajectory, are_compatible_motif_breakage_trajectories
from .obj.motif_breakage_trajectory_ensemble import MotifBreakageTrajectoryEnsemble, isinstance_motifbreakagetrajectoryensemble, are_compatible_motif_breakage_trajectory_ensembles



from .dissociation_constant import template_averaged_dissociation_constant_from_strand_reactor_parameters

from ..obj.motif_breakage_vector import MotifBreakageVector, _array_to_motif_breakage_vector
from ..obj.units import make_unit

def motif_breakage_rate_constants_from_strand_reactor_parameters(
        strand_reactor_parameters : dict,
        motiflength : int,
        alphabet : list,
        complements : list,
        standard_concentration : float = 1.
    ) -> MotifBreakageVector:
    """
    calculates (estimates for) motif_breakage_rate_constants
    from strand reactor parameters 
    """
    if motiflength != 4:
        raise NotImplementedError("function only implemented for fourmers (i.e. motiflength=4, given motiflength: {})".format(motiflength))
    breakage_rates_unit = make_unit('')

    cleavage_rate_constant = strand_reactor_parameters['r_delig']

    ligation_parameters = {}
    ligation_keys = [
        'stalling_factor_first',
        'stalling_factor_second'
    ]
    for ligation_key in ligation_keys:
        ligation_parameters[ligation_key] = strand_reactor_parameters[ligation_key]

    hybridization_parameters = {}
    hybridization_keys = [
        'dG_4_2Match_mean',
        'dG_4_1Match',
        'dG_4_0Match',
        'dG_3_1Match_mean',
        'dG_3_0Match',
        'ddG_4_2Match_alternating',
        'ddG_3_1Match_alternating'
    ]
    for hybridization_key in hybridization_keys:
        hybridization_parameters[hybridization_key] = strand_reactor_parameters[hybridization_key]

    breakage_rates = _calculate_effective_cleavage_rate_constant_for_central_breaking_tetramer(
        cleavage_rate_constant,
        standard_concentration,
        complements,
        ligation_parameters,
        hybridization_parameters
    )
    breakage_rates = _array_to_motif_breakage_vector(breakage_rates, motiflength, alphabet, breakage_rates_unit)

    breakage_rates = MotifBreakageVector(motiflength, alphabet, unit=make_unit(''))(breakage_rates)
    return breakage_rates

def _calculate_effective_cleavage_rate_constant_for_central_breaking_tetramer(
        cleavage_rate_constant,
        standard_concentration,
        complements = [1,0,3,2],
        ligation_parameters = {'stalling_factor_first' : 1.,
            'stalling_factor_second' : 1.,
            },
        hybridization_parameters = {'dG_4_2Match_mean':0.,
                    'dG_4_1Match':0.,
                    'dG_4_0Match':0.,
                    'dG_3_1Match_mean':0.,
                    'dG_3_0Match':0.,
                    'ddG_4_2Match_alternating':0.,
                    'ddG_3_1Match_alternating':0.
                    }
        ):
    template_averaged_dissociation_constant = template_averaged_dissociation_constant_from_strand_reactor_parameters(
            None,
            ligation_parameters,
            complements,
            hybridization_parameters,
            standard_concentration,
            )
    return cleavage_rate_constant * template_averaged_dissociation_constant



import numpy as np

def unzero(x, shift=1.e-12, shiftmode = 'relmin'):
    """
    function that returns an array with shift for all zero values of input array x.

    Parameters
    ----------
    x : array
    shift : float, optional
    shiftmode : string, optional
        options are 'relmin', 'relmax', 'abs'
        standing for shifting the array by
        'relmin': the minimum of x time shift
        'relmax': the maximum of x time shift
        'abs': the shift absolutely
    """
    if shiftmode == 'relmin':
        if len(x[x!=0]) == 0:
            print("Unzeroed array is completely zero.")
        else:
            print("Minimum value of unzeroed array is {}".format(np.min(x[x!=0])))
            shift = np.min(x[x!=0])*shift
    elif shiftmode == 'relmax':
        if len(x[x!=0]) == 0:
            print("Unzeroed array is completely zero.")
        else:
            print("Maximum value of unzeroed array is {}".format(np.max(x[x!=0])))
            shift = np.max(x[x!=0])*shift
    elif shiftmode == 'abs':
        shift = shift
    else:
        raise NotImplementedError("shift mode = '{}' is not implemented".format(shiftmode))
    rtrn = np.zeros(x.shape)
    rtrn[x==0.] += shift
    return rtrn



import morsaik as kdi
import numpy as np
from numpy.testing import assert_equal
import pytest

from copy import deepcopy

def _return_motif_production_category_shape_dct(motiflength : int,
        alphabet : list,
        maximum_ligation_window_length : int):
    mpcs = {}
    mpd = kdi.domains.make_motif_production_dct(alphabet,
            motiflength,
            maximum_ligation_window_length)
    for key in mpd.keys():
        mpcs[key] = mpd[key].shape
    return mpcs

def test__create_empty_motif_production_dict():
    alphabet = ['a','b','c']
    motiflength = 5
    unit = kdi.make_unit('')
    maximum_ligation_window_length = motiflength

    empv = kdi._create_empty_motif_production_dict(motiflength,
            alphabet,
            maximum_ligation_window_length
            )

    mpcsd = _return_motif_production_category_shape_dct(motiflength,
            alphabet,
            maximum_ligation_window_length
            )

    for key in mpcsd.keys():
        assert_equal(empv[key],np.zeros(mpcsd[key]))

def test_motif_production_vector():
    alphabet = ['a','b','c']
    motiflength = 5
    unit = kdi.make_unit('')
    maximum_ligation_window_length = motiflength

    motif_production_categories = kdi.domains._return_motif_production_categories(motiflength,
            alphabet,
            maximum_ligation_window_length
            )

    empv = kdi._create_empty_motif_production_dict(motiflength,
            alphabet,
            maximum_ligation_window_length
            )

    motif_production_vector = kdi.MotifProductionVector(motiflength,alphabet,unit, maximum_ligation_window_length)
    motif_production_vector = motif_production_vector(empv)

    assert(motif_production_vector.alphabet==alphabet)
    assert(motif_production_vector.number_of_letters==len(alphabet))
    assert(motif_production_vector.motiflength==motiflength)
    assert(motif_production_vector.unit==unit)
    for motif_production_category in motif_production_categories:
        assert_equal(motif_production_vector.productions[motif_production_category].val,
                empv[motif_production_category]
                )

def test_save_and_load_motif_production_vector():
    strand_trajectory_id='9999_99_99__99_99_99'
    param_file_no=0
    simulations_run_no=0
    motiflengths = [2,3,4,5]

    filepaths = kdi.utils._create_ligations_filepath_lists(strand_trajectory_id,param_file_no)[0]

    for motiflength in motiflengths:
        maximum_ligation_window_length = motiflength
        archive_path = kdi.utils.create_motif_production_trajectory_ensemble_path(
            strand_trajectory_id,
            param_file_no,
            motiflength,
            maximum_ligation_window_length
        )
        smpt = kdi.get.strand_motifs_productions_trajectory(
                motiflength,
                strand_trajectory_id,
                param_file_no=0,
                simulations_run_no=0)

        mpvd = kdi._create_empty_motif_production_dict(
            motiflength,
            smpt.alphabet,
            maximum_ligation_window_length
        )
        for key in smpt.productions.keys():
            mpvd[key] = smpt.productions[key].val[0]
        mpv = kdi.MotifProductionVector(motiflength, smpt.alphabet, kdi.make_unit('counts'), maximum_ligation_window_length)
        mpv = mpv(mpvd)
        kdi.save_motif_production_vector(archive_path, mpv)
        for key in mpv._asdict().keys():
            assert_equal(
                    mpv._asdict()[key],
                    kdi.load_motif_production_vector(archive_path)._asdict()[key]
                    )

    smpt = kdi.get.strand_motifs_productions_trajectory(6,'9999_99_99__99_99_99',0,0)
    maximum_ligation_window_length = smpt.motiflength
    actual = {}
    for key in smpt.productions.keys():
        actual[key] = deepcopy(smpt.productions[key].val[11])
    actual['length4strand_6_1_beginning'][0,1,1,0,1,0,1,0,1]=1e12
    mpv = kdi.MotifProductionVector(smpt.motiflength, smpt.alphabet, kdi.make_unit('counts'), maximum_ligation_window_length)
    actual = mpv(actual)
    # reaction:
    #a|tta-tat|at
    # in array:
    #0a|tta0-0tat|at
    #with periodic boundary conditions such that ligation spot is in the center
    # 0;0a|tta-tat|at;0
    # in indices:
    # 0;0,0|1,2,1-2,1,1|0,2;0
    assert_equal(kdi._motif_production_vector_as_array(actual)[0,0,0,1,2,1,2,1,1,0,2,0],
            1e12)



from os.path import exists

def transform_execution_time_txt_to_list(
        path : str = "archive/9999_99_99__99_99_99/data_param_file_0/4-mer_dynamics/execution_time.txt"
        ) -> list:
    with open(path,'r') as f:
        a=f.readlines()
    return [int(aa.replace('\n','').split(':')[-1]) for aa in a[1:]]

def _return_parameters_to_read_from_parameters_file() -> list:
    return ['c_ref',
            'l_critical',
            'r_delig',
            'dG_4_2Match_mean',
            'dG_4_1Match',
            'dG_4_0Match',
            'dG_3_1Match_mean',
            'dG_3_0Match',
            'ddG_4_2Match_alternating',
            'ddG_3_1Match_alternating',
            'stalling_on',
            'use_kinetic_bias_factor',
            'stalling_factor_first',
            'stalling_factor_second' 
            ]

def _create_typical_strand_reactor_dirpath(strand_trajectory_id : str) -> str:
    dirpath = './data/'
    dirpath += strand_trajectory_id + '/'
    return dirpath

def _create_typical_strand_trajectory_ensemble_dirpath(strand_trajectory_id : str,
        param_file_no : int = 0) -> str:
    dirpath = _create_typical_strand_reactor_dirpath(strand_trajectory_id)
    dirpath += 'data_exp/data_param_file_{}/'.format(param_file_no)
    return dirpath

def _create_typical_strand_trajectory_dirpath(strand_trajectory_id : str,
        param_file_no : int = 0,
        simulations_run_no : int = 0) -> str:
    dirpath = _create_typical_strand_trajectory_ensemble_dirpath(
            strand_trajectory_id,
            param_file_no)
    dirpath += 'data_simulations_run_{}/'.format(simulations_run_no)
    return dirpath

def _create_typical_strand_trajectory_section_dirpath(strand_trajectory_id : str,
        param_file_no : int = 0,
        simulations_run_no : int = 0,
        simulations_no : int = 0
        ) -> str:
    dirpath = _create_typical_strand_trajectory_dirpath(strand_trajectory_id,
            param_file_no=param_file_no,
            simulations_run_no=simulations_run_no)
    dirpath += 'data_simulations_{}/'.format(simulations_no)
    return dirpath

def _create_typical_strand_trajectory_section_complexes_filepath(strand_trajectory_id : str,
        param_file_no : int = 0,
        simulations_run_no : int = 0,
        simulations_no : int = 0,
        filename : str = 'complexes.txt'
        ) -> str:
    dirpath = _create_typical_strand_trajectory_section_dirpath(strand_trajectory_id,
            param_file_no=param_file_no,
            simulations_run_no=simulations_run_no,
            simulations_no = simulations_no
            )
    return dirpath+filename

def _create_complexes_filepath_lists(strand_trajectory_id : str,
        param_file_no :int,
        ):
    current_filepath = lambda srn, sn : _create_typical_strand_trajectory_section_complexes_filepath(
            strand_trajectory_id,
            param_file_no=param_file_no,
            simulations_run_no=srn,
            simulations_no = sn,
            )
    simulations_run_no = 0
    simulations_no = 0
    filepath_lists = []
    while exists(current_filepath(simulations_run_no,simulations_no)):
        filepaths = []
        while exists(current_filepath(simulations_run_no, simulations_no)):
            filepaths += [current_filepath(simulations_run_no, simulations_no),]
            simulations_no += 1
        filepath_lists = filepath_lists + [filepaths,]
        simulations_run_no += 1
    return filepath_lists

def _create_ligations_filepath_lists(strand_trajectory_id : str,
        param_file_no :int,
        ):
    current_filepath = lambda srn, sn : _create_typical_strand_trajectory_section_ligations_filepath(
            strand_trajectory_id,
            param_file_no=param_file_no,
            simulations_run_no=srn,
            simulations_no = sn,
            )
    simulations_run_no = 0
    simulations_no = 0
    filepath_lists = []
    while exists(current_filepath(simulations_run_no,simulations_no)):
        filepaths = []
        while exists(current_filepath(simulations_run_no, simulations_no)):
            filepaths += [current_filepath(simulations_run_no, simulations_no),]
            simulations_no += 1
        filepath_lists = filepath_lists + [filepaths,]
        simulations_run_no += 1
    return filepath_lists

def _create_typical_strand_trajectory_section_ligations_filepath(strand_trajectory_id : str,
        param_file_no : int = 0,
        simulations_run_no : int = 0,
        simulations_no : int = 0,
        filename : str = 'ligation_statistics.txt'
        ) -> str:
    dirpath = _create_typical_strand_trajectory_section_dirpath(strand_trajectory_id,
            param_file_no=param_file_no,
            simulations_run_no=simulations_run_no,
            simulations_no = simulations_no
            )
    return dirpath+filename

def _create_typical_strand_parameters_filepath(
        strand_trajectory_id : str,
        param_file_no : int = 0
    ) -> str:
    filepath = _create_typical_strand_reactor_dirpath(strand_trajectory_id)
    filepath += 'all_parameter_files/parameters_{}.txt'.format(param_file_no)
    return filepath

def _create_typical_alphabet_filepath(strand_trajectory_id : str) -> str:
    filepath = _create_typical_strand_reactor_dirpath(strand_trajectory_id)
    filepath += 'bin/initial_configurations/nucleotides.txt'
    return filepath

def read_txt(title_path, replace=[["\n", ""],["\r", ""]]):  
    """
    reads a file and returns its content as string
    """
    with open(title_path, "r", encoding ="utf8") as current_file:
        text = current_file.read()
        for rpl in replace:
            text = text.replace(rpl[0],rpl[1])
    return text



from ..read.strand_length_distribution import strand_length_distribution as read_strand_length_distribution
from ..utils.manage_strand_reactor_files import _create_typical_strand_trajectory_section_dirpath
from .alphabet import alphabet as get_alphabet
from os.path import exists

def strand_length_distribution(
        strand_trajectory_id : str,
        param_file_no : int = 0,
        ) -> dict:
    alphabet = get_alphabet(strand_trajectory_id)
    current_filepath = lambda srn, sn : _create_typical_strand_trajectory_section_dirpath(
        strand_trajectory_id,
        param_file_no=param_file_no,
        simulations_run_no=srn,
        simulations_no = sn
        ) + 'length_distribution.txt'
    simulations_run_no = 0
    simulations_no = 0
    filepath_lists = []
    while exists(current_filepath(simulations_run_no,simulations_no)):
        filepaths = []
        while exists(current_filepath(simulations_run_no, simulations_no)):
            filepaths += [current_filepath(simulations_run_no, simulations_no),]
            simulations_no += 1
        filepath_lists = filepath_lists + [filepaths,]
        simulations_run_no += 1
        simulations_no = 0
    return [read_strand_length_distribution(
        filepaths,
        alphabet
    ) for filepaths in filepath_lists]



import jax.numpy as jnp
from warnings import warn
import timeit
from functools import partial

from ..obj.times_vector import TimesVector
from ..obj.motif_vector import (MotifVector,
                                _array_to_motif_vector_dct, _motif_vector_as_array)
from ..obj.motif_trajectory import MotifTrajectory
from ..obj.motif_production_vector import (MotifProductionVector,
                                           _motif_production_vector_as_array)
from ..obj.motif_breakage_vector import (MotifBreakageVector,
                                         _motif_breakage_vector_as_array)
from .fourmer_production_rates import _shape_fprc
from .fourmer_breakage_rates import _shape_brc

from scipy.integrate import solve_ivp
from jax.experimental.ode import odeint
from diffrax import diffeqsolve
import diffrax

import timeit

from .fourmer_production_rates import compute_total_extension_rates
from .fourmer_production_rates import _set_invalid_log_rates_to_logzero
from .fourmer_production_rates import _set_invalid_production_rates_to_zero
from .fourmer_breakage_rates import fourmer_breakage_rates
from .fourmer_mass_correction import mass_correction_rates

def fourmer_trajectory_from_rate_constants(
        motif_production_rate_constants : MotifProductionVector,
        motif_production_log_rate_constants : MotifProductionVector,
        breakage_rate_constants : MotifBreakageVector,
        initial_motif_concentrations_vector : MotifVector,
        times : TimesVector,
        complements : list,
        mass_correction_rate_constant : float = 0.,
        concentrations_are_logarithmized : bool = True,
        ode_integration_method : str = 'DOP853',
        execution_time_path : str = None,
        pseudo_count_concentration : float = 1.e-12,
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None,
        first_step : float = None,
        ivp_atol : float = 1.e-6,
        ivp_rtol : float = 1.e-3,
    ) -> MotifTrajectory:
    """
    performs a fourmer reactor with the stated rate constants and initial
    conditions,
    i.e. integrates the motif ode with stated parameters.

    Parameters:
    -----------
    motif_production_rate_constants : MotifProductionVector,
    motif_production_log_rate_constants : MotifProductionVector,
    breakage_rate_constants : MotifBreakageVector,
    initial_motif_concentrations_vector : MotifVector,
    times : TimesVector,
    complements : list,
    mass_correction_rate_constant : float 
        Rate constant for compensation of numerical mass fluctuations,
        default 0.
    ode_integration_method : str
        default 'DOP853',
    execution_time_path : str
        if not None, the execution time of the ode integration is saved here
        default None,
    pseudo_count_concentration : float
        default 1.e-12,
    first_step : float
        default None
    ivp_atol : float
        atol for scipy.integrate.solve_ivp
        default 1.e-6
    ivp_rtol : float
        rtol for scipy.integrate.solve_ivp
        default 1.e-3

    Returns:
    --------
    motif_trajectory : MotifTrajectory
    """
    alphabet = initial_motif_concentrations_vector.alphabet
    motiflength = initial_motif_concentrations_vector.motiflength
    times_vector = times
    times = times.val
    if soft_reactant_threshold is None:
        soft_reactant_threshold = pseudo_count_concentration
    if motiflength!=4:
        raise ValueError("Motiflength ({}!=4) does not fit fourmer length.".format(motiflength))
    unit = initial_motif_concentrations_vector.unit
    if motiflength != 4:
        raise NotImplementedError("Motiflength needs to be four.")

    if breakage_rate_constants != 0.:
        breakage_rate_constants_array = _motif_breakage_vector_as_array(breakage_rate_constants)
    else:
        breakage_rate_constants_array = breakage_rate_constants
    #TODO assert breakage_rate_constants and motif_production_rate_constants
    # have fitting units with motif_trajectory

    initial_motif_concentrations_array = _motif_vector_as_array(initial_motif_concentrations_vector)
   
    motif_trajectory_field, times_array, execution_time = _integrate_motif_rate_equations(
        initial_motif_concentrations_array,
        number_of_letters = len(alphabet),
        motiflength=motiflength,
        complements=jnp.asarray(complements),
        concentrations_are_logarithmized = concentrations_are_logarithmized,
        fourmer_production_log_rate_constants = _motif_production_vector_as_array(motif_production_log_rate_constants),
        fourmer_production_rate_constants = _motif_production_vector_as_array(motif_production_rate_constants),
        breakage_rate_constants = breakage_rate_constants_array,
        mass_correction_rate_constant = mass_correction_rate_constant,
        t_eval = times,
        ode_integration_method = ode_integration_method,
        first_step = first_step,
        pseudo_count_concentration=pseudo_count_concentration,
        ivp_atol =ivp_atol,
        ivp_rtol = ivp_rtol,
        soft_reactant_threshold = soft_reactant_threshold,
        hard_reactant_threshold = hard_reactant_threshold
    )
    times_vector = TimesVector(times_array, times_vector.domain[0].units)
    motif_trajectory_array = motif_trajectory_field.reshape(times_array.shape+(len(alphabet)+1,)*motiflength)[:,:,1:]
    if execution_time_path is not None:
        with open(execution_time_path,'a') as f:
            f.write('\n'+str(execution_time))

    mv = MotifVector(motiflength, alphabet, unit)
    motif_vectors = [mv(_array_to_motif_vector_dct(motif_vector_array, motiflength, alphabet)) for motif_vector_array in motif_trajectory_array]
    return MotifTrajectory(motif_vectors, times_vector)

def _integrate_motif_rate_equations(
        initial_concentration_vector : jnp.ndarray,
        number_of_letters : int = 4,
        motiflength : int = 4,
        complements : list = jnp.array([1,0,3,2]),
        concentrations_are_logarithmized : bool = True,
        influx_rate_constants : jnp.ndarray = 0.,
        fourmer_production_log_rate_constants : jnp.ndarray = 0.,
        fourmer_production_rate_constants : jnp.ndarray = 1.,
        breakage_rate_constants : jnp.ndarray = 0.,
        mass_correction_rate_constant : float = 0.,
        t_eval : jnp.ndarray  = jnp.arange(0, 50000, 1),
        ode_integration_method : str = 'RK45',
        pseudo_count_concentration : float = 1.e-12,
        first_step = None,
        ivp_atol : float = 1.e-3,
        ivp_rtol : float = 1.e-6,
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None
        ):
    """
    Parameters:
    -----------
    number_of_letters : int, optional
        number of letters
        default : 4
    motiflength : int, optional
        length of the tracked motifs
        default : 4
    complements : jnp.array, optional
        array of which letters are complementary to each other
        default : jnp.array([1,0,3,2])
    concentrations_are_logarithmized : boolean, optional (default True)
        specifies whether to use logconcentrations
    initial_concentration_vector : array
        array of initial concentrations of motifs
        default : None
    fourmer_production_log_rate_constants : array
        array of effective ligation rates
        default : None
    breakage_rate_constants : float or d-array,
        breakage_rate_constants, if float is given, all breakage_rate_constants
        are assumed to be the same.
        default: 0.
    ode_integration_method : string, optional
                    solve_ivp method
                    # explicit: # 'RK45' # 'RK23' # 'DOP853'
                    # implicit: # 'Radau' # 'BDF' # 'LSODA'
                    default : 'RK45'
    ivp_atol : float
        atol for scipy.integrate.solve_ivp
        default 1.e-6
    ivp_rtol : float
        rtol for scipy.integrate.solve_ivp
        default 1.e-3

    Return:
    -------
    x_0 : NIFTy field
        solution of the integrated ODE
    """
    if motiflength != 4: 
        raise NotImplementedError(f"{motiflength=}!=4")

    if soft_reactant_threshold is None:
        soft_reactant_threshold = pseudo_count_concentration
    motif_concentration_vector_shape = (number_of_letters+1,)*(motiflength)

    if initial_concentration_vector.size == number_of_letters*(number_of_letters+1)**(motiflength-1):
        n0 = jnp.zeros(motif_concentration_vector_shape)
        n0 = n0.at[:,1:].add(
                initial_concentration_vector.reshape((number_of_letters+1,number_of_letters)+(number_of_letters+1,)*(motiflength-2))
                )
    else:
        n0 = jnp.asarray(initial_concentration_vector).copy()

    fourmer_production_log_rate_constants = _shape_fprc(
            fourmer_production_log_rate_constants,
            number_of_letters=number_of_letters,
            motiflength=motiflength,
            fprc_are_logarithmized = True
            )
    fourmer_production_rate_constants = _shape_fprc(
            fourmer_production_rate_constants,
            number_of_letters=number_of_letters,
            motiflength=motiflength,
            fprc_are_logarithmized = False
            )
    breakage_rate_constants = _shape_brc(
            breakage_rate_constants,
            number_of_letters=number_of_letters,
            motiflength=motiflength
            )

    if not concentrations_are_logarithmized:
        warn(f"motif rate equation only implemented for log concentrations to ensure positivity, will transform them to log concentrations with pseudo_count {pseudo_count_concentration} for the sake of the ode integration.")
        n0 = n0.at[n0<pseudo_count_concentration].set(pseudo_count_concentration)
        n0 = jnp.log(n0)
    n0 = n0.flatten()
    t0, t1, t_eval = _shape_t_eval(t_eval)
    rate_equations, args = _build_rate_equations(
            soft_reactant_threshold = soft_reactant_threshold,
            hard_reactant_threshold = hard_reactant_threshold,
            influx_rate_constants = influx_rate_constants,
            fourmer_production_log_rate_constants = fourmer_production_log_rate_constants,
            fourmer_production_rate_constants = fourmer_production_rate_constants,
            breakage_rate_constants = breakage_rate_constants,
            complements = complements,
            motiflength = motiflength,
            mass_correction_rate_constant = mass_correction_rate_constant,
            initial_log_concentration_array = n0.reshape(motif_concentration_vector_shape)
            )
    print(f"{args.keys() = }")
    timing_dct = {"execution_time" : None}
    timing_dct["execution_time"] = - timeit.default_timer()
    if ode_integration_method in ['RK45','RK23','DOP853','Radau','LSODA','BDF',]:
        # integrate ode
        r = solve_ivp(
                rate_equations,
                [t0,t1],
                n0,
                t_eval = t_eval,
                args = (args,),
                method = ode_integration_method,
                first_step=first_step,
                atol=ivp_atol,
                rtol=ivp_rtol,
                dense_output=True,
                )
        if not r.success:
            print(f"solve_ivp unsuccesful (probably uncomplete solution)")
            print(f"got the following message: {r.message}")
        motif_concentration_trajectory = jnp.asarray(r.y)
        times_array = jnp.asarray(r.t)
    elif ode_integration_method == 'Dopri':
        sol = odeint(
                rate_equations, #func
                n0, #y0
                t_eval, #t
                (args,), #args
                rtol=ivp_rtol,
                atol=ivp_atol,
                )
        times_array = jnp.asarray(t_eval)
        motif_concentration_trajectory = jnp.asarray(sol)
    else:
        rate_equations = diffrax.ODETerm(rate_equations)
        # Dopri5, Dopri8
        solver = diffrax.Dopri5()
        dt0 = None
        stepsize_controller = diffrax.PIDController(rtol=ivp_rtol,atol=ivp_atol)
        save_at = diffrax.SaveAt(ts=t_eval)
        r = diffeqsolve(
                rate_equations,
                solver,
                t0=t0,
                t1=t1,
                dt0=dt0,
                y0=n0,
                args = args,
                saveat = save_at,
                stepsize_controller=stepsize_controller
                )
        times_array = jnp.asarray(r.ts)
        motif_concentration_trajectory = jnp.asarray(r.ys)
    timing_dct["execution_time"] += timeit.default_timer()
    if not concentrations_are_logarithmized:
        motif_concentration_trajectory = jnp.exp(motif_concentration_trajectory)
    return jnp.asarray(motif_concentration_trajectory).T, jnp.asarray(times_array), timing_dct

def _shape_t_eval(t_eval):
    if len(t_eval)==1:
        t0=0
        t1=t_eval
        t_eval = None
    else:
        t0 = t_eval[0]
        t1 = t_eval[-1]
    return t0, t1, t_eval

def _build_rate_equations(
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None,
        **kwargs,
        ):
    """
        influx_rate_constants : jnp.ndarray = 0.,
        breakage_rate_constants : jnp.ndarray,
        fourmer_production_log_rate_constants : jnp.ndarray = 0.,
        fourmer_production_rate_constants : jnp.ndarray = 0.,
        complements : list = jnp.array([1,0,3,2]),
        motiflength : int = 4,
    """
    print(f"{kwargs.keys() = }")
    kwargs.setdefault('influx_rate_constants', 0.)
    kwargs.setdefault('fourmer_production_log_rate_constants', 0.)
    kwargs.setdefault(' fourmer_production_rate_constants',  0.)
    kwargs.setdefault('breakage_rate_constants',0.)
    kwargs.setdefault('complements', jnp.array([1,0,3,2]))
    kwargs.setdefault('motiflength', 4)
    kwargs.setdefault('mass_correction_rate_constant', 0.)
    print(f"{kwargs.keys() = }")
    if kwargs['mass_correction_rate_constant'] != 0.:
        if 'initial_log_concentration_array' not in kwargs.keys():
            raise TypeError("For mass_correction_rate_constant != 0, initial_log_concentration_array needs to be specified")
    rate_equations = _fourmer_rate_equations(kwargs,
                                             soft_reactant_threshold = soft_reactant_threshold,
                                             hard_reactant_threshold = hard_reactant_threshold)
    return rate_equations, kwargs

def _fourmer_production_equations(
        t, y, arg1,
        soft_reactant_threshold : float = 0.,
        hard_reactant_threshold : float = None
        ):
    """
    arg1 = ([influx_rate_constants,
             (fourmer_production_log_rate_constants,fourmer_production_rate_constants),
             complements,
             motiflength],)

    """
    leerc = arg1['fourmer_production_log_rate_constants']
    eerc=arg1['fourmer_production_rate_constants']
    complements = arg1['complements']
    motiflength = arg1['motiflength']
    if eerc.shape[0] == eerc.shape[1]:
        eerc = eerc[:,1:,1:,:,:,1:,1:,:]
    if leerc.shape[0] == leerc.shape[1]:
        leerc = leerc[:,1:,1:,:,:,1:,1:,:]
    return jnp.asarray(compute_total_extension_rates(
        jnp.asarray(y.reshape((len(complements)+1,)*motiflength)),
        log_rate_constants = leerc,
        rate_constants = eerc,
        motiflength = 4,
        number_of_letters = eerc.shape[1],
        soft_reactant_threshold = soft_reactant_threshold,
        hard_reactant_threshold = hard_reactant_threshold
        )).reshape(-1)

def _fourmer_influx_equations(t, y, arg):
    return arg['influx_rate_constants'].reshape(-1)*jnp.exp(-y).reshape(-1)

def _fourmer_breakage_equations(
        t, y, arg2,
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None
        ):
    """
    arg2 : dict
        specify 'breakage_rate_constants', 'complements', 'motiflength'
    soft_reactant_threshold : float
        minimal concentration of a reactant to contribute fully to a reaction,
        for smaller concentrations the reaction rates get damped or clipped
        default : 0., i.e. clipping deactivated
    hard_reactant_threshold : float
       concentration of a reactant at which all its reactions are set to zero.
       Between 'soft_reactant_threshold' and 'hard_reactant_threshold' the reaction rates smoothly
       transition between fully contribution of the reactant and clipping
       via a cos-funtion.
       Default : None, setting hard_reactant_threshold to soft_reactant_threshold/2
    """
    if not isinstance(arg2,dict):
        raise TypeError("arg2 supposed to be dictionary")
    arg2.setdefault('breakage_rate_constants',0.)
    arg2.setdefault('complements', jnp.array([1,0,3,2]))
    arg2.setdefault('motiflength', 4)
    breakage_rate_constants = arg2['breakage_rate_constants']
    complements = arg2['complements']
    motiflength = arg2['motiflength']

    breakage_logc_diff = fourmer_breakage_rates(
            y.reshape((complements.size+1,)*motiflength),
            effective_breakage_rate_constants = breakage_rate_constants,
            soft_reactant_threshold = soft_reactant_threshold,
            hard_reactant_threshold = hard_reactant_threshold
            )
    return breakage_logc_diff.reshape(-1)

def _fourmer_mass_correction_rates(t,y, arg3):
    initial_log_concentration_array = arg3['initial_log_concentration_array']
    mass_correction_rate_constant = arg3['mass_correction_rate_constant']
    complements = arg3['complements']
    motiflength = arg3['motiflength']
    return mass_correction_rates(
            initial_log_concentration_array,
            y.reshape((complements.size+1,)*motiflength),
            weight = mass_correction_rate_constant).reshape(-1)

def _fourmer_rate_equations(
        kwargs,
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None
        ):
    """
    returns a function: the rate equation with arguments
    t, y, args : dict; 
    args = [influx_rate_constants,
             fourmer_production_log_rate_constants,
             fourmer_production_rate_constants,
             breakage_rate_constants,
             complements,
             motiflength,
             mass_correction_rate_constant,
             initial_log_concentration_array
             ],
    """
    exclude_influx =  jnp.all(jnp.asarray(kwargs['influx_rate_constants']) == 0.)
    include_influx = 1-exclude_influx
    if exclude_influx:
        print("Influx rate constants equal 0, thus, influx turned off.")
    exclude_breakage = jnp.all(kwargs['breakage_rate_constants'] == 0.)
    include_breakage = 1-exclude_breakage
    if exclude_breakage:
        print("Breakage rate constant equals 0, thus, breakage turned off.")
    exclude_extension = jnp.all(kwargs['fourmer_production_rate_constants']==0.)
    include_extension = 1-exclude_extension
    if exclude_extension:
        print("Extension rate constants equal 0, thus, fourmer extension turned off.")
    exclude_mass_correction = bool(kwargs['mass_correction_rate_constant']==0.)
    include_mass_correction = 1-exclude_mass_correction
    if exclude_mass_correction:
        print("Mass correction rate constant equals 0, thus, mass correction turned off.")
    def fre(t,y,args):
        freturn = jnp.zeros(y.shape)
        if include_influx:
            freturn = freturn + _fourmer_influx_equations(t, y, args)
        if include_breakage:
            freturn = freturn + _fourmer_breakage_equations(t, y, args,
                                                            soft_reactant_threshold=soft_reactant_threshold,
                                                            hard_reactant_threshold=hard_reactant_threshold)
        if include_extension:
            freturn = freturn + _fourmer_production_equations(t, y, args,
                                                              soft_reactant_threshold=soft_reactant_threshold,
                                                              hard_reactant_threshold=hard_reactant_threshold)
        if include_mass_correction:
            freturn = freturn + _fourmer_mass_correction_rates(t, y, args)
        return freturn
    return fre



import matplotlib.pyplot as plt
import numpy as np
from itertools import product as iterprod

from ..obj.motif_trajectory_ensemble import MotifTrajectoryEnsemble
from ..obj.units import transform_unit_to_str
from .colors import _choose_color

from ..infer.zebraness import system_level_motif_zebraness as infer_system_level_motif_zebraness
from ..infer.zebraness import individual_motif_zebraness as infer_individual_motif_zebraness

def zebraness(
        motif_trajectory_ensemble : MotifTrajectoryEnsemble,
        linestyle = '-',
        color = 'b',
        axis = None
    ):
    if axis is None:
        axis = plt.gca()
    ii = 0
    for motif_trajectory in motif_trajectory_ensemble.trajectories:
        colr = _choose_color(color, ii/np.max([1,len(motif_trajectory_ensemble.trajectories)-1]))
        print("Plotting zebraness...")
        times = motif_trajectory.times
        motifs = motif_trajectory.motifs.val

        zebras = motif_trajectory.motifs['length2strand'].val[:,0,1]+motif_trajectory.motifs['length2strand'].val[:,1,0]
        c0 = np.sum(motif_trajectory.motifs['length2strand'].val.reshape((times.shape[0],-1)), axis = 1)
        zebraness = zebras/c0

        axis.plot(times.val, zebraness, linestyle = linestyle, color = colr)
        axis.set_xlabel('Time ' + transform_unit_to_str(times.domain[0].units))
        axis.set_ylabel('Zebraness []')
        axis.set_ylim([0.,1.])
        ii += 1

def nest_mean(rough_list : list) -> np.ndarray:
    """
    arranges list of arrays such that all start at zeroth position and gives back mean of the occupied values.
    """
    # even list
    nested_mean_length = np.max([len(rl) for rl in rough_list])
    nested_mean = []
    for rl in rough_list:
        rl = np.array(list(rl) + [np.nan]*(nested_mean_length-len(rl)))
        nested_mean += [rl]
    return np.nanmean(nested_mean, axis = 0)

def system_level_motif_zebraness(
        motif_trajectory_ensembles : list,
        plotting_time_windows : list = None,
        plot_mean : list = [False],
        plot_parameters : list = [{
            'linestyle' : '-',
            'color' : 'b',
            'alpha' : 0.3,
            'label' : 'Strand',
        }],
        plotpath : str = None,
        plotformats : list = ['.pdf'],
        annotation : dict = {
            'text':'',
            'xy':(0,1),
            'xycoords':'axes fraction',
            'xytext':(+0.5,-0.5),
            'textcoords':'offset fontsize',
            'verticalalignment':'top',
            'bbox':{
                'facecolor':'white',
                'edgecolor':'none',
                'alpha':0.7,
            },
            },
    ):
    """
    plots system level motif zebraness, i.e. the fraction of alternating 2-mers
    in the whole motif reactor.

    Parameters:
    -----------
    motif_trajectory_ensemble : list of MotifTrajectoryEnsemble
    plotting_time_windows : list = None,
    plot_mean : list
        default: [False]
    plot_parameters : list
        of dicts of format {
            'linestyle' : '-',
            'color' : 'b',
            'alpha' : 0.3,
            'label' : 'Strand',
        }
    plotpath : str
        default: None,
    plotformats : list
        default: ['.pdf'],
    annotation : dict
        default: {
        'text':'',
        'xy':(0,1),
        'xycoords':'axes fraction',
        'xytext':(+0.5,-0.5),
        'textcoords':'offset fontsize',
        'verticalalignment':'top',
        'bbox':{
            'facecolor':'white',
            'edgecolor':'none',
            'alpha':0.7,
        },
        },

    Returns:
    --------
    None

    """
    if not isinstance(motif_trajectory_ensembles,list):
        motif_trajectory_ensembles = [motif_trajectory_ensembles,]
    if not isinstance(plot_parameters, list):
        plot_parameters = [plot_parameters,]
    times_list = [
            motif_trajectory_ensemble.trajectories[
                np.argmax([motif_trajectory_ensemble.trajectories[ti].times.size
                           for ti in range(len(motif_trajectory_ensemble.trajectories))])].times
                for motif_trajectory_ensemble in motif_trajectory_ensembles
                ]
    motif_trajectories_list = [
            motif_trajectory_ensemble.trajectories
            for motif_trajectory_ensemble in motif_trajectory_ensembles
                          ]

    if plotting_time_windows is None:
        plotting_time_windows = [[times_list[0].val[0],times_list[0].val[-1]]]

    for tptt in plotting_time_windows:
        for yscale, xscale in iterprod(['linear','log'],repeat=2):
            plt.close('all')
            for ii in range(len(motif_trajectories_list)):
                motif_trajectories = motif_trajectories_list[ii]
                colrs = plot_parameters[ii]['color']
                trj_times_val = motif_trajectories[0].times.val
                if plot_mean[ii]:
                    zebraness_mean = []
                for jj in range(len(motif_trajectories)):
                    colr = _choose_color(colrs, jj/np.max([1,len(motif_trajectories)-1]))
                    trj = motif_trajectories[jj]
                    if len(trj.times.val)>len(trj_times_val):
                        trj_times_val = trj.times.val
                    zebraness = infer_system_level_motif_zebraness(trj, axis=1)
                    if plot_mean[ii]:
                        zebraness_mean += [zebraness]
                    plt.plot(trj.times.val,
                            zebraness,
                            plot_parameters[ii]['linestyle'],
                            color = colr,
                            label= plot_parameters[ii]['label'],
                            alpha  = plot_parameters[ii]['alpha'])
                if plot_mean[ii]:
                    zebraness_mean = nest_mean(zebraness_mean)
                    colr = _choose_color(color, color_parameter = len(colrs//2))
                    plt.plot(trj_times_val,
                            zebraness_mean,
                            '-',
                            color = colr,
                            label= plot_parameters[ii]['label'],
                            alpha  = np.min((1.,plot_parameters[ii]['alpha']*2.))
                            )
            plt.annotate(**annotation)
            plt.xscale(xscale)
            plt.yscale(yscale)
            plt.xlabel('Time ' + transform_unit_to_str(times_list[0].domain[0].units))
            plt.ylabel('Zebraness []')
            if tptt[0] is None:
                tptt[0] = 10**int(np.log10(times_list[0].val[1]))
            if tptt[1] is None:
                tptt[1] = np.max([tl.val[-1] for tl in times_list])
            if xscale=='log' and tptt[0]==0.:
                tptt[0] = 10**int(np.log10(times_list[0].val[1]))
            plt.xlim(*tptt)
            plt.ylim([0.,1.])
            if plotpath is not None:
                pp = plotpath+'{title}_{tptt0}-{tptt1}_{yscale}-{xscale}'
                figname = pp.format(
                    title='motif_zebraness',
                    tptt0=str(tptt[0]),
                    tptt1=str(tptt[1]),
                    xscale=xscale,
                    yscale=yscale,
                )
                for plotformat in plotformats:
                    plt.savefig(figname + plotformat)
                print('plotted {}'.format(figname))
                plt.close()



from os.path import exists

from .alphabet import alphabet as get_alphabet
from ..read.strand_motifs_trajectory import strand_motifs_trajectory as read_strand_motifs_trajectory
from ..utils.manage_strand_reactor_files import _create_typical_strand_trajectory_section_complexes_filepath 

def strand_motifs_trajectory(
        motiflength : int,
        strand_trajectory_id : str,
        param_file_no : int = 0,
        simulations_run_no : int = 0,
        skiprows : int = 2,
        execution_time_path : str = None
        ):
    alphabet = get_alphabet(strand_trajectory_id)

    current_filepath = lambda sn : _create_typical_strand_trajectory_section_complexes_filepath(
            strand_trajectory_id,
            param_file_no=param_file_no,
            simulations_run_no=simulations_run_no,
            simulations_no = sn,
            )

    simulations_no = 0
    filepaths = []
    while exists(current_filepath(simulations_no)):
        filepaths += [current_filepath(simulations_no),]
        simulations_no += 1
    if execution_time_path is not None:
        with open(execution_time_path,'a') as f:
            f.write('\n'+str(simulations_no))
    return read_strand_motifs_trajectory(filepaths,
            alphabet,
            motiflength,
            skiprows =skiprows)



import numpy as np
import nifty8 as ift
from nifty8 import MultiDomain, DomainTuple
from nifty8.utilities import check_object_identity, frozendict, indent

from warnings import warn

from typing import Union

from .hamming_space import HammingSpace
from .motif_space import MotifSpace
from ..obj.units import make_unit, Unit
from .motif_space import _return_motif_categories

def MotifHyphenSpace(alphabet : list,
                         motiflength : int,
                         units : Unit,
                         ) -> MotifSpace:
    ms = MotifSpace.make(alphabet, motiflength, units=units, monomers_included = False)
    dct = {}
    for key in ms.keys():
        for hyphen_index in range(1,ms[key][0].wordlength):
            hyphen_key = key + '_{}'.format(hyphen_index)
            dct[hyphen_key] = ms[key][0]
    return ift.MultiDomain.make(dct)

class MotifHyphenSpace(MotifSpace):
    """
    Domain of Motif Hyphens,
    like motif breakage rate constants,
    ligations projected onto motif space.
    """

    @staticmethod
    def make(alphabet, motiflength, units : Unit = None):
        """Creates a MultiDomain object from a dictionary of names and domains

        Parameters
        ----------
        inp : MultiDomain or dict{name: DomainTuple}
            The already built MultiDomain or a dictionary of DomainTuples

        Returns
        ------
        A MultiDomain with the input Domains as domains
        """
        if units is None:
            units = make_unit('bits',np.log2(len(alphabet)))
        else:
            units = make_unit(units)
        if (units/make_unit('bits', np.log2(len(alphabet))))!=1.:
            warn("Length of alphabet suggests that units of MotifSpace is {}".format(make_unit('bits', np.log2(len(alphabet))))+'but given units are {}.'.format(units))
        if isinstance(alphabet,int):
            alphabet = list(1 + np.arange(alphabet))
        inp = make_hyphen_dct(alphabet, motiflength)

        if isinstance(inp, MotifHyphenSpace):
            return inp
        if not isinstance(inp, dict):
            raise TypeError("dict expected")
        tmp = {}
        for key, value in inp.items():
            if not isinstance(key, str):
                raise TypeError("keys must be strings")
            tmp[key] = DomainTuple.make(value)
        tmp = frozendict(tmp)
        obj = MotifHyphenSpace._domainCache.get(tmp)
        if obj is not None:
            return obj
        obj = MotifHyphenSpace(tmp,
                alphabet=alphabet,
                motiflength=motiflength,
                units=units,
                _callingfrommake=True)
        MotifHyphenSpace._domainCache[tmp] = obj
        return obj
    @property
    def motiflength(self):
        return self._motiflength
    @property
    def alphabet(self):
        return self._alphabet
    @property
    def number_of_letters(self):
        return len(self._alphabet)
    @property
    def units(self):
        return self._units

def _hyphen_id(motif_category : str,
        hyphen_spot : int) -> str:
    """
    The hyphen id is given by the categories of the motif
    and hyphen spot.
    """
    return motif_category + '_{}'.format(hyphen_spot)

def _determine_motif_categories_and_hyphen_spots(motiflength : int,
        ) -> Union[list,list]:
    if motiflength < 4:
        # Hyphen spot goes for motiflength = 2 only one value for continuations, for beginnings and endings no value
        # for motiflength = 3 for ending no value (already considered in continuation),
        # for beginning and for continuation one value
        hyphen_spots = np.array([motiflength])-2

        motif_categories = _return_motif_categories()[-2-int(motiflength>2):-1]
    else:
        """
        all central hyphens are considered no matter the category, additionally
        - for strands: all hyphen spots
        - for beginnings: all hyphens before the central hyphen
        - for ends: all hyphens after the central hyphen
        """
        ligation_spots = np.arange(1,motiflength-2)

        product_categories = [product_category.format(ligation_window_length-2) for product_category in _return_motif_categories()]
    return product_categories, template_categories, ligation_spots


def make_hyphen_dct(alphabet : list,
                          motiflength : int,
                          ) -> dict:
    """
    Creates a motif hyphen dictionary.
    Parameters:
    -----------
    alphabet : list
    motiflength : int,

    Returns:
    --------
    motif_hyphen_dct : dict
    """
    hyphen_dct = {}
    hamming = lambda ii : HammingSpace(alphabet,ii)
    motif_categories = _return_motif_categories(motiflength=motiflength)
    # completely captured strands (1-mers don't have hyphens)
    for strandlength in range(2,motiflength-1):
        motif_category = motif_categories[:-3][strandlength-1]
        hyphen_ids =[_hyphen_id(motif_category,hyphen_spot) for hyphen_spot in np.arange(1,strandlength)]
        for hyphen_key in hyphen_ids:
            hyphen_dct[hyphen_key] = hamming(strandlength)
    # beginnings
    motif_category = motif_categories[-3]
    hyphen_ids =[_hyphen_id(motif_category,hyphen_spot) for hyphen_spot in np.arange(1,motiflength-motiflength//2)]
    for hyphen_key in hyphen_ids:
        hyphen_dct[hyphen_key] = hamming(motiflength-1)
    # continuations
    motif_category = motif_categories[-2]
    hyphen_ids = [_hyphen_id(motif_category, motiflength-motiflength//2)]
    for hyphen_key in hyphen_ids:
        hyphen_dct[hyphen_key] = hamming(motiflength)
    # ends
    motif_category = motif_categories[-1]
    hyphen_ids =[_hyphen_id(motif_category,hyphen_spot) for hyphen_spot in np.arange(motiflength-motiflength//2,motiflength-1)]
    for hyphen_key in hyphen_ids:
        hyphen_dct[hyphen_key] = hamming(motiflength-1)
        for key in hyphen_dct.keys():
            print(key)
    return hyphen_dct



import jax.numpy as jnp
from warnings import warn

from jax import config
config.update("jax_enable_x64", True)

from ._rates_utils import _set_invalid_logc_to_log0
from ._rates_utils import _set_invalid_logc_diff_to_zero
from ._rates_utils import _cut_low_concentrations
from ._smooth_rate_clipping import _clip_smoothly

def _initialize_empty_fourmer_production_rates(
        number_of_letters : int = 4,
        motiflength : int = 4
        ) -> jnp.ndarray :
    if motiflength != 4:
        raise NotImplementedError("motiflength!=4")
    shape = (number_of_letters+1,)*(motiflength-1)
    shape = shape*2+(number_of_letters+1,)*motiflength
    return jnp.zeros(shape)

def _set_invalid_production_rates_to_zero(
        reaction_rates : jnp.ndarray,
        ) -> jnp.ndarray:
    rr = reaction_rates.at[:,0].set(0)
    rr = rr.at[:,:,0].set(0)
    rr = rr.at[:,:,:,:,:,0].set(0)
    return rr.at[:,:,:,:,:,:,0].set(0)

def _set_invalid_log_production_rates_to_logzero(
        log_reaction_rates : jnp.ndarray,
        pseudo_count_concentration : float = 1.e-12
        ) -> jnp.ndarray:
    log0 = jnp.log(pseudo_count_concentration)
    rr = log_reaction_rates.at[:,:,0].set(log0)
    rr = rr.at[:,0].set(log0)
    rr = rr.at[:,:,0].set(log0)
    rr = rr.at[:,:,:,:,:,0].set(log0)
    return rr.at[:,:,:,:,:,:,0].set(log0)

def _set_invalid_rates_to_zero(
        reaction_rates : jnp.ndarray,
        ) -> jnp.ndarray:
    """
    set invalid rates to zeros.
    Format of the rates: [l0,l1,l2,r1,r2,r3,t0,t1,t2,t3]
    """
    rr = reaction_rates.at[:,:,0].set(0)
    rr = rr.at[1:,0].set(0)
    rr = rr.at[:,:,:,0].set(0)
    rr = rr.at[:,:,:,:,0,1:].set(0)
    rr = rr.at[:,:,:,:,:,:,:,0,:,:].set(0)
    return rr.at[:,:,:,:,:,:,:,:,0,:].set(0)

def _set_invalid_log_rates_to_logzero(
        log_reaction_rates : jnp.ndarray,
        pseudo_count_concentration : float = 1.e-12
        ) -> jnp.ndarray:
    log0 = jnp.log(pseudo_count_concentration)
    lrr = log_reaction_rates.at[:,:,0].set(log0)
    lrr = lrr.at[1:,0].set(log0)
    lrr = lrr.at[:,:,:,0].set(log0)
    lrr = lrr.at[:,:,:,:,0,1:].set(log0)
    lrr = lrr.at[:,:,:,:,:,:,:,0,:,:].set(log0)
    return lrr.at[:,:,:,:,:,:,:,:,0,:].set(log0)

def _add_log_rate_constants(
        log_rate_constants : jnp.ndarray,
        log_reaction_rates : jnp.ndarray,
        pseudo_count_concentration : float = 1.e-12

        ) -> jnp.ndarray:
    return log_reaction_rates.at[:,:,1:,1:,:,:,:,1:,1:,:].add(log_rate_constants.reshape((1,)+log_rate_constants.shape[:len(log_rate_constants.shape)//2]+(1,)+log_rate_constants.shape[len(log_rate_constants.shape)//2:]))

def _multiply_rr_with_rc(rr,rc):
    return rr.at[:,:,1:,1:,:,:,:,1:,1:,:].multiply(rc[None,:,:,:,:,None])

def _add_template_logc(
        fourmer_logc : jnp.ndarray,
        log_reaction_rates : jnp.ndarray,
        number_of_letters : int = 4,
        pseudo_count_concentration : float = 1.e-12
        ) -> jnp.ndarray:
    motiflength = 4
    fourmer_logc = _set_invalid_logc_to_log0(fourmer_logc, pseudo_count_concentration/10.)
    return log_reaction_rates.at[:,:,:,:,:,:,:,1:,1:,:].add(fourmer_logc[:,1:,1:,:].reshape((1,)*(2*(motiflength-1))+(number_of_letters+1,number_of_letters,number_of_letters,number_of_letters+1)))

def _add_extended_end_logc(
        fourmer_logc : jnp.ndarray,
        log_reaction_rates : jnp.ndarray,
        number_of_letters : int = 4,
        pseudo_count_concentration : float = 1.e-12
        ) -> jnp.ndarray:
    motiflength = 4
    fourmer_logc = _set_invalid_logc_to_log0(fourmer_logc, pseudo_count_concentration/10.)
    # add monomers
    lrr = log_reaction_rates.at[0,0,1:].add(fourmer_logc[0,1:,0,0].reshape((number_of_letters,)+(1,)*(motiflength-1)+(1,)*motiflength))
    # add ending motifs
    return lrr.at[:,1:,1:].add(
            fourmer_logc[:,1:,1:,0].reshape(
                (number_of_letters+1,number_of_letters,number_of_letters)+(1,)*(motiflength-1)+(1,)*motiflength
                )
            )

def _add_extending_beginning_logc(
        fourmer_logc : jnp.ndarray,
        log_reaction_rates : jnp.ndarray,
        motiflength : int = 4,
        pseudo_count_concentration : float = 1.e-12
        ) -> jnp.ndarray:
    if motiflength != 4:
        raise NotImplementedError("motiflength!=4")
    fourmer_logc = _set_invalid_logc_to_log0(fourmer_logc, pseudo_count_concentration/10.)
    # monomers (0100) and dimers (0110)
    lrr = log_reaction_rates.at[:,:,:,1:,:,0].add(
            fourmer_logc[0,1:,:,0].reshape(
                (1,)*(motiflength-1)+fourmer_logc[0,1:,:,0].shape+(1,)*(motiflength)
            )
            )
    # beginnings 0111
    lrr = lrr.at[:,:,:,1:,1:,1:].add(
            fourmer_logc[0,1:,1:,1:].reshape(
                (1,)*(motiflength-1)+fourmer_logc[0,1:,1:,1:].shape+(1,)*(motiflength)
            )
            )
    return lrr

def _subtract_produced_fourmers(
        fourmer_logc : jnp.ndarray,
        log_reaction_rates : jnp.ndarray,
        product_index : int = 0,
        motiflength : int = 4,
        pseudo_count_concentration : float  = 1.e-12
        ) -> jnp.ndarray:
    if motiflength != 4:
        raise NotImplementedError("motiflength != 4")
    # only subtract product, where term is nonzero (so greater log0)
    log0 = jnp.log(pseudo_count_concentration)
    if product_index==0:
        lrr = log_reaction_rates.at[:,0,:].set(log0-1)
    elif product_index==2:
        lrr = log_reaction_rates.at[:,:,:,:,0].set(log0-1)
    else:
        lrr = log_reaction_rates
    fourmer_logc = _set_invalid_logc_to_log0(fourmer_logc, pseudo_count_concentration/10.)
    nonzero = lrr > log0
    produced_fourmer_logc = _set_invalid_logc_diff_to_zero(fourmer_logc, print_warning = False)
    produced_fourmer_logc = produced_fourmer_logc.reshape(
                (1,)*product_index + produced_fourmer_logc.shape + (1,)*(2*(motiflength-1)-product_index)
                )
    produced_fourmer_logc = nonzero*produced_fourmer_logc
    lrr = lrr.at[:].subtract(
            produced_fourmer_logc
            )
    return _set_invalid_log_rates_to_logzero(lrr, pseudo_count_concentration/10.)

def _exponentiate_log_reaction_rates(
        log_reaction_rates : jnp.ndarray
        ) -> jnp.ndarray:
    return jnp.exp(log_reaction_rates)

def _compute_extended_end_motif_reaction_logc_rates(
        fourmer_logc : jnp.ndarray,
        log_rate_constants : jnp.ndarray = 0.,
        rate_constants : jnp.ndarray = 1.,
        motiflength : int = 4,
        number_of_letters : int = 4,
        pseudo_count_concentration : float = 1.e-12,
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None,
        ):
    if soft_reactant_threshold is None:
        soft_reactant_threshold = pseudo_count_concentration
    if hard_reactant_threshold is None:
        hard_reactant_threshold = soft_reactant_threshold/2.
    fourmer_logc = _set_invalid_logc_to_log0(fourmer_logc, pseudo_count_concentration/10.)
    log_rate_constants, rate_constants = _split_and_shape_fprc(
            (log_rate_constants, rate_constants),
            number_of_letters=number_of_letters,
            motiflength=motiflength,
            pseudo_count_concentration=pseudo_count_concentration
            )
    lrr = _initialize_empty_fourmer_production_rates(number_of_letters)
    lrr = _add_log_rate_constants(log_rate_constants,lrr)
    lrr = _set_invalid_log_rates_to_logzero(lrr, pseudo_count_concentration/10.)
    lrr = _add_template_logc(fourmer_logc,lrr, number_of_letters, pseudo_count_concentration)
    lrr = _add_extending_beginning_logc(fourmer_logc,lrr)
    rr = _exponentiate_log_reaction_rates(lrr)
    # clip smoothly
    rr = _clip(fourmer_logc, rr, soft_reactant_threshold, hard_reactant_threshold)
    rr = _multiply_rr_with_rc(rr,rate_constants)
    return -_set_invalid_rates_to_zero(rr)

def _clip(fourmer_logc, rr, soft_reactant_threshold, hard_reactant_threshold):
    # check for extensive or excessive format
    if rr.shape[2]<fourmer_logc.shape[1]:
        fourmer_logc = fourmer_logc[:,1:]
    elif rr.shape[2]>fourmer_logc.shape[1]:
        fourmer_logc = (hard_reactant_threshold*jnp.ones(
                (fourmer_logc.shape[0],)*len(fourmer_logc.shape)
                )).at[:,1:].set(
                        fourmer_logc
                )
    weight = _clip_smoothly(
            jnp.exp(fourmer_logc),
            soft_reactant_threshold=soft_reactant_threshold,
            hard_reactant_threshold=hard_reactant_threshold)
    return weight*rr

def _compute_extending_beginning_motif_reaction_logc_rates(
        fourmer_logc : jnp.ndarray,
        log_rate_constants : jnp.ndarray = 0.,
        rate_constants : jnp.ndarray = 1.,
        motiflength : int = 4,
        number_of_letters : int = 4,
        pseudo_count_concentration : float = 1.e-12,
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None,
        ) -> jnp.ndarray :
    if soft_reactant_threshold is None:
        soft_reactant_threshold = pseudo_count_concentration
    if hard_reactant_threshold is None:
        hard_reactant_threshold = soft_reactant_threshold/2.
    fourmer_logc = _set_invalid_logc_to_log0(fourmer_logc, pseudo_count_concentration/10.)
    log_rate_constants, rate_constants = _split_and_shape_fprc(
            (log_rate_constants, rate_constants),
            number_of_letters=number_of_letters,
            motiflength=motiflength,
            pseudo_count_concentration=pseudo_count_concentration
            )
    lrr = _initialize_empty_fourmer_production_rates(number_of_letters)
    lrr = _add_log_rate_constants(log_rate_constants,lrr)
    lrr = _set_invalid_log_rates_to_logzero(lrr, pseudo_count_concentration/10.)
    lrr = _add_template_logc(fourmer_logc, lrr, number_of_letters, pseudo_count_concentration)
    lrr = _add_extended_end_logc(fourmer_logc, lrr, number_of_letters)
    rr = _exponentiate_log_reaction_rates(lrr)
    # clip smoothly
    rr = _clip(fourmer_logc, rr, soft_reactant_threshold, hard_reactant_threshold)
    rr = _multiply_rr_with_rc(rr,rate_constants)
    return -_set_invalid_rates_to_zero(rr)

def _compute_produced_motif_reaction_logc_rates(
        fourmer_logc : jnp.ndarray,
        log_rate_constants : jnp.ndarray = 0.,
        rate_constants : jnp.ndarray = 1.,
        product_index : int = 0,
        number_of_letters : int = 4,
        motiflength : int = 4,
        pseudo_count_concentration : float = 1.e-12,
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None,
        ) -> jnp.ndarray:
    if soft_reactant_threshold is None:
        soft_reactant_threshold = pseudo_count_concentration
    if hard_reactant_threshold is None:
        hard_reactant_threshold = soft_reactant_threshold/2.
    fourmer_logc = _set_invalid_logc_to_log0(fourmer_logc, pseudo_count_concentration/10.)
    log_rate_constants, rate_constants = _split_and_shape_fprc(
            (log_rate_constants, rate_constants),
            number_of_letters=number_of_letters,
            motiflength=motiflength,
            pseudo_count_concentration=pseudo_count_concentration
            )
    lrr = _initialize_empty_fourmer_production_rates(number_of_letters)
    lrr = _add_log_rate_constants(log_rate_constants, lrr)
    lrr = _set_invalid_log_rates_to_logzero(lrr, pseudo_count_concentration/10.)
    lrr = _add_template_logc(fourmer_logc,lrr,number_of_letters, pseudo_count_concentration)
    lrr = _add_extended_end_logc(fourmer_logc,lrr, number_of_letters)
    lrr = _add_extending_beginning_logc(fourmer_logc,lrr)
    lrr = _subtract_produced_fourmers(
            fourmer_logc,
            lrr,
            product_index=product_index
            )
    rr = _exponentiate_log_reaction_rates(lrr)
    # clip smoothly
    rr = _clip(fourmer_logc, rr, soft_reactant_threshold, hard_reactant_threshold)
    rr = _multiply_rr_with_rc(rr, rate_constants)
    return +_set_invalid_rates_to_zero(rr)

def compute_logc_diff_extended_end_motifs(
        fourmer_logc : jnp.ndarray,
        log_rate_constants : jnp.ndarray = 0.,
        rate_constants : jnp.ndarray = 1.,
        motiflength : int = 4,
        number_of_letters : int = 4,
        pseudo_count_concentration : float = 1.e-12,
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None
        ) -> jnp.ndarray:
    log_rate_constants, rate_constants = _split_and_shape_fprc(
            (log_rate_constants, rate_constants),
            number_of_letters=number_of_letters,
            motiflength=motiflength,
            pseudo_count_concentration=pseudo_count_concentration
            )
    fourmer_logc = _set_invalid_logc_to_log0(fourmer_logc, pseudo_count_concentration/10.)
    if motiflength!=4:
        raise NotImplementedError("motiflength!=4")
    logc_diff = jnp.zeros((number_of_letters+1,)*motiflength)
    # monomers
    extended_end_motif_reaction_rates = _compute_extended_end_motif_reaction_logc_rates(
            fourmer_logc,
            log_rate_constants,
            rate_constants,
            motiflength,
            number_of_letters,
            pseudo_count_concentration,
            soft_reactant_threshold=soft_reactant_threshold,
            hard_reactant_threshold=hard_reactant_threshold
            )
    logc_diff = logc_diff.at[0,1:,0,0].add(
            jnp.sum(
                extended_end_motif_reaction_rates[0,0,1:],
                axis = tuple([ii for ii in range(1,len(extended_end_motif_reaction_rates[0,0,1:].shape))]
                             ),
                )
            )
    # motifs
    logc_diff = logc_diff.at[:,1:,1:,0].add(
            jnp.sum(extended_end_motif_reaction_rates[:,1:,1:],
                axis = tuple([ii for ii in range((motiflength-1),len(extended_end_motif_reaction_rates[:,1:,1:].shape))])
                    )
                )
    return _set_invalid_logc_diff_to_zero(logc_diff, print_warning = False)

def compute_logc_diff_extending_beginning_motifs(
        fourmer_logc : jnp.ndarray,
        log_rate_constants : jnp.ndarray = 0.,
        rate_constants : jnp.ndarray = 1.,
        motiflength : int = 4,
        number_of_letters : int = 4,
        pseudo_count_concentration : float = 1.e-12,
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None
        ) -> jnp.ndarray:
    fourmer_logc = _set_invalid_logc_to_log0(fourmer_logc, pseudo_count_concentration/10.)
    logc_diff = jnp.zeros((number_of_letters+1,)*motiflength)
    log_rate_constants, rate_constants = _split_and_shape_fprc(
            (log_rate_constants, rate_constants),
            number_of_letters=number_of_letters,
            motiflength=motiflength,
            pseudo_count_concentration=pseudo_count_concentration
            )
    extending_beginning_motif_reaction_rates = _compute_extending_beginning_motif_reaction_logc_rates(
            fourmer_logc,
            log_rate_constants,
            rate_constants,
            motiflength,
            number_of_letters,
            pseudo_count_concentration,
            soft_reactant_threshold=soft_reactant_threshold,
            hard_reactant_threshold=hard_reactant_threshold
            )
    # monomers and dimers
    logc_diff = logc_diff.at[0,1:,:,0].add(
            jnp.sum(extending_beginning_motif_reaction_rates[:,:,:,1:,:,0],
                axis = tuple([ii for ii in range(motiflength-1)])+tuple(
                    [ii for ii in range(2*motiflength-2-1,len(extending_beginning_motif_reaction_rates.shape)-1)])
                    )
                )
    # beginnings
    logc_diff = logc_diff.at[0,1:,1:,1:].add(
            jnp.sum(extending_beginning_motif_reaction_rates[:,:,:,1:,1:,1:],
                axis = tuple([ii for ii in range(motiflength-1)])+tuple([ii for ii in range(2*motiflength-2,len(extending_beginning_motif_reaction_rates.shape))])
                    )
                )
    return  _set_invalid_logc_diff_to_zero(logc_diff, print_warning = False)

def compute_logc_diff_produced_motifs(
        fourmer_logc : jnp.ndarray,
        log_rate_constants : jnp.ndarray = 0.,
        rate_constants : jnp.ndarray = 1.,
        motiflength : int = 4,
        number_of_letters : int = 4,
        pseudo_count_concentration : float = 1.e-12,
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None
        ) -> jnp.ndarray:
    fourmer_logc = _set_invalid_logc_to_log0(fourmer_logc, pseudo_count_concentration/10.)
    logc_diff = jnp.zeros((number_of_letters+1,)*motiflength)
    log_rate_constants, basic_rate_constants = _split_and_shape_fprc(
            (log_rate_constants, rate_constants),
            number_of_letters=number_of_letters,
            motiflength=motiflength,
            pseudo_count_concentration=pseudo_count_concentration
            )
    for product_index in range(motiflength-1):
        produced_motif_reaction_rates =  _compute_produced_motif_reaction_logc_rates(
                fourmer_logc,
                log_rate_constants,
                basic_rate_constants,
                product_index,
                number_of_letters,
                motiflength=motiflength,
                pseudo_count_concentration=pseudo_count_concentration,
                soft_reactant_threshold=soft_reactant_threshold,
                hard_reactant_threshold=hard_reactant_threshold
                )
        pmrr = jnp.sum(
                produced_motif_reaction_rates,
                axis = tuple(
                    [ii for ii in range(product_index)]) + tuple(
                        [ii for ii in range(motiflength+product_index,3*motiflength-2)])
                    )
        # dimers, beginning, ends and continuations
        logc_diff = logc_diff.at[:,1:,1:,:].add(pmrr[:,1:,1:,:])
    return _set_invalid_logc_diff_to_zero(logc_diff, print_warning = False)

def compute_total_extension_rates(
        fourmer_logc : jnp.ndarray,
        log_rate_constants : jnp.ndarray = 0.,
        rate_constants : jnp.ndarray = 1.,
        motiflength : int = 4,
        number_of_letters : int = 4,
        pseudo_count_concentration : float = 1.e-12,
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None,
        conserve_mass : bool = False,
        fourmer_logc_slope : float = 0.,
        logc_diff_slope : float = 0.
        ) -> jnp.ndarray:
    """

    Parameters:
    -----------

    conserve_mass : bool
        whether concentration of monomers shall compensate variation in the
        total mass (deprecated! Please use mass_correction_rates for this from
        now on, since this feature leads to inconsistencies in the zebraness
        and strand number)
        default False

    Returns:
    --------
    total_extension_rates : jnp.ndarray
    """
    fourmer_logc = _set_invalid_logc_to_log0(fourmer_logc, pseudo_count_concentration/10.)
    log_rate_constants, basic_rate_constants = _split_and_shape_fprc(
        (log_rate_constants, rate_constants),
        number_of_letters=number_of_letters,
        motiflength=motiflength,
        pseudo_count_concentration=pseudo_count_concentration
        )
    # initialize logc_diff
    logc_diff = jnp.zeros((number_of_letters+1,)*motiflength)
    # consume_left_reactant(self):
    logc_diff = jnp.add(logc_diff,
            compute_logc_diff_extended_end_motifs(
                fourmer_logc,
                log_rate_constants,
                rate_constants,
                motiflength,
                number_of_letters,
                pseudo_count_concentration,
                soft_reactant_threshold=soft_reactant_threshold,
                hard_reactant_threshold=hard_reactant_threshold
                )
            )
    # consume_right_reactant(self):
    logc_diff = jnp.add(logc_diff,
            compute_logc_diff_extending_beginning_motifs(
                fourmer_logc,
                log_rate_constants,
                rate_constants,
                motiflength,
                number_of_letters,
                pseudo_count_concentration,
                soft_reactant_threshold=soft_reactant_threshold,
                hard_reactant_threshold=hard_reactant_threshold
                )
            )
    # create_products
    logc_diff = jnp.add(logc_diff,
            compute_logc_diff_produced_motifs(
                fourmer_logc,
                log_rate_constants,
                rate_constants,
                motiflength,
                number_of_letters,
                pseudo_count_concentration,
                soft_reactant_threshold=soft_reactant_threshold,
                hard_reactant_threshold=hard_reactant_threshold
                )
            )
    if conserve_mass:
        c_diff = jnp.exp(fourmer_logc)*logc_diff
        c_diff = c_diff.at[0,1:,0,0].set(-jnp.sum(c_diff[:,1:,1:,:],axis=(0,2,3))-jnp.sum(c_diff[:,1:,1:,0],axis=(0,1)))
        logc_diff = jnp.exp(-fourmer_logc)*c_diff
    return _set_invalid_logc_diff_to_zero(logc_diff, print_warning = False)

def compute_motif_extensions(
        fourmer_logc : jnp.ndarray,
        log_rate_constants : jnp.ndarray = 0.,
        basic_rate_constants : jnp.ndarray = 1.,
        motiflength : int = 4,
        number_of_letters : int = 4,
        pseudo_count_concentration : float = 1.e-12,
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None
        ) -> jnp.ndarray:
    if soft_reactant_threshold is None:
        soft_reactant_threshold = pseudo_count_concentration
    if hard_reactant_threshold is None:
        hard_reactant_threshold = soft_reactant_threshold/2.
    fourmer_logc = _set_invalid_logc_to_log0(fourmer_logc, pseudo_count_concentration/10.)
    log_rate_constants, basic_rate_constants = _split_and_shape_fprc(
            (log_rate_constants, basic_rate_constants),
            number_of_letters=number_of_letters,
            motiflength=motiflength,
            pseudo_count_concentration=pseudo_count_concentration
            )
    lrr = _initialize_empty_fourmer_production_rates(number_of_letters)
    lrr = _add_log_rate_constants(log_rate_constants, lrr)
    lrr = _set_invalid_log_rates_to_logzero(lrr, pseudo_count_concentration/10.)
    lrr = _add_extended_end_logc(fourmer_logc,lrr, number_of_letters)
    lrr = _add_extending_beginning_logc(fourmer_logc,lrr)
    lrr = _add_template_logc(fourmer_logc,lrr, number_of_letters, pseudo_count_concentration)
    rr = _exponentiate_log_reaction_rates(lrr)
    #rr = rr.at[rr<=pseudo_count_concentration].set(0.)
    rr = _clip(fourmer_logc, rr, soft_reactant_threshold, hard_reactant_threshold)
    rr = _multiply_rr_with_rc(rr,basic_rate_constants)
    return jnp.sum(rr, axis=(0,2*(motiflength-1)-1))

def _split_and_shape_fprc(
        lrc_brc,
        number_of_letters = 4,
        motiflength = 4,
        pseudo_count_concentration = 1.e-12,
        ):
    lrc, brc = lrc_brc
    lrc = _shape_fprc(
            lrc,
            number_of_letters = number_of_letters,
            motiflength = motiflength,
            pseudo_count_concentration = pseudo_count_concentration,
            fprc_are_logarithmized=True,
            shape_is_thin=True
            )
    brc = _shape_fprc(
            brc,
            number_of_letters = number_of_letters,
            motiflength = motiflength,
            pseudo_count_concentration = pseudo_count_concentration,
            fprc_are_logarithmized=False,
            shape_is_thin=True
            )
    return lrc,brc

def _shape_fprc(
        fourmer_production_rate_constants,
        number_of_letters : int = 4,
        motiflength : int = 4,
        pseudo_count_concentration : float = 1.e-12,
        fprc_are_logarithmized : bool = False,
        shape_is_thin : bool = False
        ) -> jnp.ndarray:
    motif_production_rate_constants_shape = (number_of_letters+1,)*(2*motiflength)
    if isinstance(fourmer_production_rate_constants,(float,int)):
        eerc = float(fprc_are_logarithmized)*jnp.ones(motif_production_rate_constants_shape)
        fourmer_production_rate_constants = eerc.at[:,1:,1:,:,:,1:,1:,:].set(fourmer_production_rate_constants)
    elif fourmer_production_rate_constants.size == (number_of_letters+1)**motiflength*number_of_letters**motiflength:
        eerc_new = float(fprc_are_logarithmized)*jnp.ones(motif_production_rate_constants_shape)
        fourmer_production_rate_constants = eerc_new.at[:,1:,1:,:,:,1:,1:,:].set(fourmer_production_rate_constants)
    if fourmer_production_rate_constants.shape != motif_production_rate_constants_shape:
        raise ValueError(f"{fourmer_production_rate_constants.shape=} != {motif_production_rate_constants_shape=}")
    if fprc_are_logarithmized:
        if shape_is_thin:
            #thin_shape = (number_of_letters+1,number_of_letters)
            #thin_shape += shape[::-1]
            #thin_shape += shape
            return fourmer_production_rate_constants[:,1:,1:,:,:,1:,1:,:]
        else:
            return _set_invalid_log_production_rates_to_logzero(
                    fourmer_production_rate_constants,
                    pseudo_count_concentration=pseudo_count_concentration
                    )
    else:
        if shape_is_thin:
            return fourmer_production_rate_constants[:,1:,1:,:,:,1:,1:,:]
        else:
            return _set_invalid_production_rates_to_zero(
                    fourmer_production_rate_constants,
                    )



from os.path import exists

from ..read.strand_motifs_productions_trajectory import strand_motifs_productions_trajectory as read_strand_motifs_productions_trajectory
from .alphabet import alphabet as get_alphabet
from ..utils.manage_strand_reactor_files import _create_typical_strand_trajectory_section_ligations_filepath 

def strand_motifs_productions_trajectory(
        motiflength : int,
        strand_trajectory_id : str,
        param_file_no : int = 0,
        simulations_run_no : int = 0,
        skiprows : int = 2,
        maximum_ligation_window_length : int = None
        ):
    if maximum_ligation_window_length is None:
        maximum_ligation_window_length = motiflength

    alphabet = get_alphabet(strand_trajectory_id)

    current_filepath = lambda sn : _create_typical_strand_trajectory_section_ligations_filepath(
            strand_trajectory_id,
            param_file_no=param_file_no,
            simulations_run_no=simulations_run_no,
            simulations_no = sn,
            )

    simulations_no = 0
    filepaths = []
    while exists(current_filepath(simulations_no)):
        filepaths += [current_filepath(simulations_no),]
        simulations_no += 1

    return read_strand_motifs_productions_trajectory(filepaths, alphabet, motiflength,
            maximum_ligation_window_length)
